#!/usr/bin/env python3
"""
A simple tool for joining and filtering tables
"""
import argparse
import sys

import pandas as pd

from pybiotk.utils import read_table, write_table, ignore


def main(table_list, outfile, namefile, noheader, column, exclude=False, contains=False):
    df_list = []
    for table in table_list:
        header = None if noheader else 0
        df = read_table(table, header=header, dtype=str, comment="#")
        if namefile is not None:
            names = set(j for i in namefile for j in i.split())
            if exclude:
                if contains:
                    df = df.loc[~df.iloc[:, column].str.contains("|".join(names))]
                else:
                    df = df.loc[~df.iloc[:, column].isin(names)]
            else:
                if contains:
                    df = df.loc[df.iloc[:, column].str.contains("|".join(names))]
                else:
                    df = df.loc[df.iloc[:, column].isin(names)]
        df_list.append(df)
    if len(df_list) == 1:
        out_df = df_list[0]
    else:
        out_df = pd.concat(df_list)
    write_table(out_df, outfile, header=not noheader)


@ignore
def run():
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("input", type=str, nargs="+",
                        help="Input tables.")
    parser.add_argument('-o', dest='output', type=argparse.FileType('w'),
                        default=sys.stdout, help="output file name [stdout]")
    parser.add_argument('-n', dest="namefile", type=argparse.FileType('r'), default=(None if sys.stdin.isatty() else sys.stdin),
                        help="whose name is listed in FILE|stdin")
    parser.add_argument('-H', "--noheader", dest="noheader", action="store_true", help="if noheader")
    parser.add_argument('-c', dest="column", type=int, default=0, help="name column")
    parser.add_argument('-e', dest="exclude", action="store_true", help="name is not listed in FILE|stdin")
    parser.add_argument('--contains', dest="contains", action="store_true", help="contains one of a substrings listed in FILE|stdin.")
    args = parser.parse_args()
    main(args.input, args.output, args.namefile, args.noheader, args.column, args.exclude, args.contains)


if __name__ == "__main__":
    run()



#!/usr/bin/env python3
import argparse
import os
import re
import sys
from collections import namedtuple
from typing import Optional, Sequence, TextIO, Union

import pandas as pd

from pybiotk.utils import write_table, ignore
from stream import cat, head


def parse_cutadapt(filename: str):
    total_line = ""
    reads1_line = ""
    reads2_line = ""
    for line in cat(filename) | head(15):
        line = line.strip()
        if line.startswith("Total read pairs processed:"):
            total_line = line.replace(",", "")
        if line.startswith("Read 1 with adapter:"):
            reads1_line = line.replace(",", "")
        if line.startswith("Read 2 with adapter:"):
            reads2_line = line.replace(",", "")
    assert all([total_line, reads1_line, reads2_line])
    total = re.compile('Total read pairs processed:\\s*(\\d*)').findall(total_line)[0]
    read1 = re.compile('Read 1 with adapter:\\s*(\\d*.*)').findall(reads1_line)[0]
    read1_num = read1.split(" ")[0]
    read1_percent = read1.split(" ")[1].lstrip("(").rstrip(")")
    read2 = re.compile('Read 2 with adapter:\\s*(\\d*.*)').findall(reads2_line)[0]
    read2_num = read2.split(" ")[0]
    read2_percent = read2.split(" ")[1].lstrip("(").rstrip(")")

    cutadapt_summary = namedtuple("cutadapt_summary", [
        "total_read_pairs",
        "read1_with_adapter",
        "read2_with_adapter",
        "read1_with_adapter_percent",
        "read2_with_adapter_percent"
        ])
    return cutadapt_summary(total, read1_num, read2_num, read1_percent, read2_percent)


def parse_bowtie2(filename: str):
    lines = [line.strip() for line in cat(filename)]
    input_read_pairs = lines[0].split()[0]
    mapped_read_pairs = str(int(lines[3].split()[0]) + int(lines[4].split()[0]))
    alignment_rate = lines[5].split()[0]
    bowtie2_summary = namedtuple("bowtie2_summary", [
        "input_read_pairs",
        "mapped_read_pairs",
        "alignment_rate"
    ])

    return bowtie2_summary(input_read_pairs, mapped_read_pairs, alignment_rate)


def parse_picards_rmdup(filename: str):
    lines = ""
    for line in cat(filename) | head(10):
        line = line.strip()
        if line.startswith("Unknown Library"):
            lines = line
    assert lines
    line_toks = lines.split()
    estimated_librarysize = line_toks[-1]
    duplication_rate = f"{float(line_toks[-2]) * 100:.2f}%"
    picards_rmdup_summary = namedtuple("picards_rmdup_summary", ["duplication_rate", "estimated_librarysize"])
    return picards_rmdup_summary(duplication_rate, estimated_librarysize)


def parse_flagstat(filename: str):
    lines = [line.strip() for line in cat(filename)]
    filtered_reads = lines[0].split()[0]
    return filtered_reads


def parse_sample_name(sample_name: str):
    rep_finder = re.compile('.*(\\d)$').findall(sample_name)
    rep = f"rep{rep_finder[0]}" if rep_finder else "rep0"
    group = sample_name.rstrip(rep).rstrip("_").rstrip("-")
    sample = namedtuple("sample", ["name", "replicate", "group"])
    return sample(sample_name, rep, group)


def parse_path(filename: str):
    return os.path.splitext(os.path.basename(filename))[0].split(".")[0]


def main(filepath_or_buffer: Union[str, TextIO],
         cutadapt: Optional[Sequence[str]],
         bowtie2: Optional[Sequence[str]] = None,
         picards_rmdup: Optional[Sequence[str]] = None,
         flagstat: Optional[Sequence[str]] = None,
         reads_in_peaks: Optional[Sequence[str]] = None,
         sample_names: Optional[Sequence[str]] = None,
         replicates: Optional[Sequence[str]] = None,
         group: Optional[Sequence[str]] = None,
         ):
    sample_names = sample_names if sample_names is not None else [parse_path(i) for i in cutadapt]
    group = group if group is not None else [parse_sample_name(i).group for i in sample_names]
    replicates = replicates if replicates is not None else [parse_sample_name(i).replicate for i in sample_names]
    assert len(sample_names) == len(cutadapt)
    data = dict()
    data["samples"] = sample_names
    data["group"] = group
    data["replicates"] = replicates
    total_read_pairs = []
    read1_with_adapter = []
    read2_with_adapter = []
    read1_with_adapter_percent = []
    read2_with_adapter_percent = []
    for log in cutadapt:
        cutadapt_summary = parse_cutadapt(log)
        total_read_pairs.append(cutadapt_summary.total_read_pairs)
        read1_with_adapter.append(cutadapt_summary.read1_with_adapter)
        read2_with_adapter.append(cutadapt_summary.read2_with_adapter)
        read1_with_adapter_percent.append(cutadapt_summary.read1_with_adapter_percent)
        read2_with_adapter_percent.append(cutadapt_summary.read2_with_adapter_percent)
    data["total_read_pairs"] = total_read_pairs
    data["read1_with_adapter"] = read1_with_adapter
    data["read2_with_adapter"] = read2_with_adapter
    data["read1_with_adapter_percent"] = read1_with_adapter_percent
    data["read2_with_adapter_percent"] = read2_with_adapter_percent
    if bowtie2 is not None:
        assert len(bowtie2) == len(sample_names)
        input_read_pairs = []
        mapped_read_pairs = []
        alignment_rate = []
        for log in bowtie2:
            bowtie2_log = parse_bowtie2(log)
            input_read_pairs.append(bowtie2_log.input_read_pairs)
            mapped_read_pairs.append(bowtie2_log.mapped_read_pairs)
            alignment_rate.append(bowtie2_log.alignment_rate)
        data["input_read_pairs"] = input_read_pairs
        data["mapped_read_pairs"] = mapped_read_pairs
        data["alignment_rate"] = alignment_rate
    if picards_rmdup is not None:
        assert len(picards_rmdup) == len(sample_names)
        duplication_rate = []
        estimated_librarysize = []
        for log in picards_rmdup:
            picards_rmdup_summary = parse_picards_rmdup(log)
            duplication_rate.append(picards_rmdup_summary.duplication_rate)
            estimated_librarysize.append(picards_rmdup_summary.estimated_librarysize)
        data["duplication_rate"] = duplication_rate
        data["estimated_librarysize"] = estimated_librarysize
    if flagstat is not None:
        assert len(flagstat) == len(sample_names)
        filtered_reads = []
        for log in flagstat:
            filtered_reads.append(parse_flagstat(log))
        data["filtered_reads"] = filtered_reads
    if reads_in_peaks is not None:
        assert len(reads_in_peaks) == len(sample_names)
        peaks = []
        in_peaks = []
        for log in reads_in_peaks:
            with open(log) as l:
                toks = l.read().split()
            peaks.append(toks[0])
            in_peaks.append(toks[1])
        data["peaks_count"] = peaks
        data["reads_in_peaks"] = in_peaks
    df = pd.DataFrame(data)
    write_table(df, filepath_or_buffer)


@ignore
def run():
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-o', dest='output', type=argparse.FileType('w'),
                        default=sys.stdout, help="output file name.")
    parser.add_argument('-s', "--sample_names", dest="sample_names", nargs="+",
                        type=str, default=None, help="sample_names.")
    parser.add_argument('-r', "--replicates", dest="replicates", nargs="+",
                        type=str, default=None, help="replicates.")
    parser.add_argument('-g', "--group", dest="group", nargs="+",
                        type=str, default=None, help="group.")
    parser.add_argument('--cutadapt', dest="cutadapt", nargs="+", type=str,
                        required=True, help="cutadapt log.")
    parser.add_argument('--bowtie2', dest="bowtie2", nargs="+", type=str,
                        default=None, help="bowtie2 log.")
    parser.add_argument('--picards_rmdup', dest="picards_rmdup", nargs="+", type=str,
                        default=None, help="picards_rmdup.")
    parser.add_argument('--flagstat', dest="flagstat", nargs="+", type=str,
                        default=None, help="flagstat.")
    parser.add_argument('--reads_in_peaks', dest="reads_in_peaks", nargs="+", type=str,
                        default=None, help="reads_in_peaks.")
    args = parser.parse_args()
    main(args.output, args.cutadapt, args.bowtie2, args.picards_rmdup, args.flagstat, args.reads_in_peaks, args.sample_names, args.replicates, args.group)


if __name__ == "__main__":
    run()



#!/usr/bin/env python3
import argparse
import time
from typing import Sequence

from pybiotk.io import FastaFile
from pybiotk.utils import logging, ignore


def main(filename, chromList: Sequence[str], wrap: bool = True):
    logging.info("reading fasta ....")
    start = time.perf_counter()
    with FastaFile(filename) as fa:
        fa.stdout(referenceList=chromList, wrap=wrap)
    end = time.perf_counter()
    logging.info(f"task finished in {end-start:.2f}s.")


@ignore
def run():
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument(dest="fasta", type=str, help="Genome fasta file.")
    parser.add_argument("-c", dest="chromlist", nargs="+", default=None, help="chrom list for split genome.")
    parser.add_argument("--wrap", dest="wrap", action="store_true", help="line-wrapped display.")
    args = parser.parse_args()
    main(args.fasta, args.chromlist, args.wrap)


if __name__ == "__main__":
    run()



from __future__ import annotations

import itertools
import shlex
import subprocess
import sys
from collections import deque
from enum import Enum
from typing import (
    Set,
    List,
    Any,
    Generic,
    Iterable,
    Iterator,
    Tuple,
    TypeVar,
    Callable,
    Literal,
    Optional,
    overload
)

__all__ = [
    "stream_type",
    "stream",
]

_T = TypeVar("_T")
_S = TypeVar("_S")


class stream_type(Enum):
    CMD_IN = 0
    STD_IN = 1
    FILE_IN = 2


class _stream_of:
    @staticmethod
    def sys_in(cmd: str, shell: bool = False) -> Iterator[str]:
        """single cmd;no linux pip
        """
        cmd2 = None
        if shell:
            cmd2 = cmd
        else:
            cmd2 = shlex.split(cmd)
        with subprocess.Popen(cmd2, stdout=subprocess.PIPE, encoding='utf8', shell=shell) as proc:
            assert proc.stdout is not None
            for line in proc.stdout:
                line = line.strip()
                if line:
                    yield line

    @staticmethod
    def std_in() -> Iterator[str]:
        for line in sys.stdin:
            line = line.strip()
            if line:
                yield line

    @staticmethod
    def file_in(file: str) -> Iterator[str]:
        with open(file) as f:
            for line in f:
                line = line.strip()
                if line:
                    yield line


def it_window(iterables: Iterable[_T], window: int = 2, step: int = 1) -> Iterator[Tuple[_T, ...]]:
    it = iter(iterables)
    d = deque(itertools.islice(it, window), maxlen=window)
    n = 0
    yield tuple(d)
    for x in it:
        d.append(x)
        n += 1
        if n == step:
            yield tuple(d)
            n = 0
    if n != 0 and step > n:
        for _ in range(step-n):
            d.popleft()
        if d:
            yield tuple(d)


def _groupby(it: Iterable[_T], fun: Optional[Callable[[_T, _T], bool]] = None, compare: Literal['start', 'end'] = 'start') -> Iterator[Tuple[_T, ...]]:
    # it must be sorted before
    assert fun is not None
    it = iter(it)
    x = deque(itertools.islice(it, 1))
    compared = 0 if compare == 'start' else -1
    for i in it:
        if fun(x[compared], i):
            x.append(i)
        else:
            yield tuple(x)
            x.clear()
            x.append(i)
    if x:
        yield tuple(x)


class stream(Generic[_T]):

    def __init__(self, iterables: Iterable[_T]):
        self.it = iterables
        self.is_consumed = False

    def filter(self, func: Callable[[_T], bool]):
        self.it = filter(func, self.it)
        return self

    def flatten(self):
        return self.transform(itertools.chain.from_iterable)

    def concat(self, it: Iterable[_T]):
        self.it = itertools.chain(self.it, it)
        return self

    def chain(self):
        return self.transform(itertools.chain)

    def head(self, n: int = 10):
        self.it = itertools.islice(self.it, n)
        return self

    def abandon(self, n: int):
        self.it = itertools.islice(self.it, n, None)
        return self

    def tail(self, n: int = 10):
        self.it = iter(deque(self.it, maxlen=n))
        return self

    def move(self) -> Iterable[_T]:
        return self.it

    def list(self) -> List[_T]:
        return list(self.move())

    def set(self) -> Set[_T]:
        return set(self.move())

    def tuple(self) -> Tuple[_T, ...]:
        return tuple(self.move())

    def collect(self, kind: Literal['list', 'dict', 'tuple', 'set']):
        return eval(kind)(self.move())

    def to_file(self, name: str,  **kargs):
        with open(name, 'wt', **kargs) as f:
            for i in self.move():
                f.write(f"{i}\n")

    def transform(self, func: Callable[[Iterable[_T]], Iterable[_S]], *args, **kargs) -> stream[_S]:
        return stream._transform(self, func, *args, **kargs)

    def map(self, func: Callable[[_T], _S]) -> stream[_S]:
        return stream._map(self, func)

    def round(self, func: Callable[..., bool], b: int = 1, a: int = 1):
        return self.window(b+a+1).filter(lambda x: func(x[b])).flatten()

    def peek(self, func: Callable[[_T], Any]):
        """debug
        """
        def inner(it: Iterable[_T]) -> Iterable[_T]:
            for i in it:
                try:
                    func(i)
                    yield i
                except Exception:
                    raise RuntimeError('{}'.format(i))
        self.it = inner(self.it)
        return self

    def print(self):
        self.for_each(print)

    def for_each(self, fun: Callable[[_T], Any]) :
        assert self.is_consumed is False
        self.is_consumed = True
        for i in self.it:
            fun(i)

    @overload
    def count(self, fun: Callable[[int], _S]) -> _S: ...

    @overload
    def count(self) -> int: ...

    def count(self, fun=None):
        value = sum(map(lambda x: 1, self.it))
        if fun is not None:
            return fun(value)
        else:
            return value

    def nothing(self, *args, **kargs):
        return self

    def window(self, window: int = 2, step: int = 1) -> stream[Tuple[_T, ...]]:
        assert window > 0 and step > 0
        return stream(it_window(self.it, window=window, step=step))

    def groupby(self, key=None):
        return self.transform(stream._groupby, key=key)

    def groupby2(self, fun: Callable[[_T, _T], bool], compare: Literal['start', 'end'] = 'start'):
        """
        Parameters:
        ----------
        compare: where the behind element compare to in one group

        Example:
        >>> a=stream([1,1,2,2,3,3,4,5,4,6,7,8]).groupby2(lambda x,y: y-x<2, compare='end').print()
        >>> (1, 1, 2, 2, 3, 3, 4, 5, 4)
        >>> (6, 7, 8)
        >>> a=stream([1,1,2,2,3,3,4,5,4,6,7,8]).groupby2(lambda x,y: y-x<2).print()
        >>> (1, 1, 2, 2)
        >>> (3, 3, 4)
        >>> (5, 4, 6)
        >>> (7, 8)
        """
        return self.transform(_groupby, fun=fun, compare=compare)

    def sort(self, key: Optional[Callable] = None, reverse: bool = False):
        """sort a small number of element

        Parameters
        ----------
        key : Callable, optional
            the value to be sorted, by default None
        reverse : bool, optional
            ascending , by default False
        """
        self.it: Iterable[_T] = iter(sorted(self.it, key=key, reverse=reverse))
        return self

    def uniq(self, key: Optional[Callable] = None):
        """apply to a sorted iter, remove repeat item.
        Args:
            key : Callable, Defaults to None.
                func that gets a object values.

        Examples:
        --------
        >>> a = [1, 1, 2, 3, 3 , 4, 4, 5, 4, 4]
        >>> stream(iter(a)).uniq().for_each(print)
        ... 1
        ... 2
        ... 3
        ... 4
        ... 5
        ... 4
        """
        self.it = (list(y)[0] for _, y in itertools.groupby(self.it, key=key))
        return self

    @staticmethod
    def _map(x: stream, func: Callable[[_T], _S]) -> stream[_S]:
        x.it = map(func, x.it)
        return x

    @staticmethod
    def _transform(x: stream, func: Callable[[Iterable[_T]], Iterable[_S]], *args, **kargs) -> stream[_S]:
        x.it = iter(func(x.it, *args, **kargs))
        return x

    @staticmethod
    def _groupby(x, key=None) -> Iterator[Tuple[_T, ...]]:
        it = (tuple(g) for _, g in itertools.groupby(x, key=key))
        return it

    @staticmethod
    def of(type: stream_type, s: Optional[str] = None, shell: bool = False) -> stream[str]:
        if type is stream_type.STD_IN:
            return stream(_stream_of.std_in())
        else:
            assert s is not None
            if type is stream_type.CMD_IN:
                return stream(_stream_of.sys_in(s, shell=shell))
            else:
                return stream(_stream_of.file_in(s))



#!/usr/bin/env python3
import argparse
import os
import sys
from typing import Optional

import numpy as np
import pysam

from pybiotk.io import count_bam_size
from pybiotk.utils import logging, ignore


def main(filename: str = "-", output: str = "-", bam: bool = False, bamsize: Optional[int] = None, count: int = 10000):
    mode = "wb" if bam else "w"
    alignment = []
    header = None
    if bamsize is None:
        logging.info("bamsize is unknown, it will take some time and memory to calculate ...")
        if filename == "-":
            with pysam.AlignmentFile(filename) as alignmentfile:
                header = alignmentfile.header
                alignment = list(alignmentfile)
                bamsize = len(alignment)
        else:
            bamsize = count_bam_size(filename)
    
    index_arr = np.arange(bamsize)
    np.random.shuffle(index_arr)
    select = set(index_arr[:count])
    
    if not alignment:
        alignment = pysam.AlignmentFile(filename)
        header = alignment.header
        
    with pysam.AlignmentFile(output, mode, header=header) as outf:
        for index, segment in enumerate(alignment):
            if index in select:
                outf.write(segment)


@ignore
def run():
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument(dest="filename", nargs='?', type=str, default=(None if sys.stdin.isatty() else "-"), help="the input BAM file, must with header [stdin]")
    parser.add_argument("-o", "--output", dest="file", type=str, default="-", help="Write output to FILE [stdout]")
    parser.add_argument("-b", "--bam", dest="bam", action="store_true", help="Output BAM")
    parser.add_argument("--bamsize", dest="bamsize", type=int, default=None, help="bamsize. If unknown, it will take some time and memory to calculate")
    parser.add_argument("-n", dest="count", type=int, default=10000, help="number of alignments to grab")
    args = parser.parse_args()
    if args.filename is None:
        args = parser.parse_args(['-h'])
    if not args.filename == "-" and not os.path.exists(args.filename):
        raise OSError(f"No such file or directory: {args.filename}")
    try:
        main(args.filename, args.file, args.bam, args.bamsize, args.count)
    except ValueError:
        logging.error("An error occurs, input BAM file must have a header, use '-h' option when using samtools view.")
        raise


if __name__ == "__main__":
    run()



#!/usr/bin/env python3
import argparse
import sys
from typing import Sequence, Optional, TextIO

from pybiotk.io import GtfFile
from pybiotk.utils import ignore
from stream import write


def main(istream: TextIO,
         features: Optional[Sequence[str]] = None,
         gene_types: Optional[Sequence[str]] = None,
         transcript_types: Optional[Sequence[str]] = None,
         output: Optional[TextIO] = None,
         transcript_ids: Optional[Sequence[str]] = None,
         transcript_names: Optional[Sequence[str]] = None,
         gene_ids: Optional[Sequence[str]] = None,
         gene_names: Optional[Sequence[str]] = None
         ):
    with GtfFile(istream) as gtf:
        gtf.iter(features=features,
                 gene_types=gene_types,
                 transcript_types=transcript_types,
                 transcript_ids=transcript_ids,
                 transcript_names=transcript_names,
                 gene_ids=gene_ids,
                 gene_names=gene_names
                 ) | write(output)


@ignore
def run():
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('gtf', nargs='?', type=argparse.FileType('r'),
                        default=(None if sys.stdin.isatty() else sys.stdin),
                        help="gtf file download from Genecode, or a sorted gtf file.")
    parser.add_argument('-f', '--features', dest='features', default=['exon'], nargs="+", help="annotation features.")
    parser.add_argument('-g', '--gene_types', dest='gene_types', nargs="+",
                        default=None, help="choose gene types to filter gtf.")
    parser.add_argument('-t', '--transcript_types', dest='transcript_types', nargs="+",
                        default=None, help="choose transcript types to filter gtf.")
    parser.add_argument('-o', dest='output', type=argparse.FileType('w'),
                        default=sys.stdout, help="output file name.")
    parser.add_argument('--transcript_ids', dest='transcript_ids', nargs="+",
                        default=None, help="choose transcript ids to filter gtf.")
    parser.add_argument('--transcript_names', dest='transcript_names', nargs="+",
                        default=None, help="choose transcript names to filter gtf.")
    parser.add_argument('--gene_ids', dest='gene_ids', nargs="+",
                        default=None, help="choose gene ids to filter gtf.")
    parser.add_argument('--gene_names', dest='gene_names', nargs="+",
                        default=None, help="choose gene names to filter gtf.")
    args = parser.parse_args()
    if args.gtf is None:
        parser.parse_args(['-h'])

    main(args.gtf, args.features, args.gene_types, args.transcript_types, args.output,
         args.transcript_ids, args.transcript_names, args.gene_ids, args.gene_names)


if __name__ == "__main__":
    run()



# -*- coding: utf-8 -*-
from typing import List, Dict, Iterable, Iterator

from pybiotk.bx.bitset import BinnedBitSet, MAX


def binned_bitsets_from_list(lst: Iterable[List]) -> Dict:
    """Read a list into a dictionary of bitsets"""
    last_chrom = None
    last_bitset = None
    bitsets = dict()
    for bed in lst:
        chrom = bed[0]
        if chrom != last_chrom:
            if chrom not in bitsets:
                bitsets[chrom] = BinnedBitSet(MAX)
            last_chrom = chrom
            last_bitset = bitsets[chrom]
        start, end = int(bed[1]), int(bed[2])
        last_bitset.set_range(start, end - start)
    return bitsets


def unionBed3(lst: Iterable[List]) -> Iterator[List]:
    """Take the union of 3 column bed files. return a new list"""
    bitsets = binned_bitsets_from_list(lst)
    for chrom in bitsets:
        bits = bitsets[chrom]
        end = 0
        while 1:
            start = bits.next_set(end)
            if start == bits.size:
                break
            end = bits.next_clear(start)
            yield [chrom, start, end]
    del bitsets


def intersectBed3(lst1, lst2) -> Iterator[List]:
    """Take the intersection of two bed files (3 column bed files)"""
    bits1 = binned_bitsets_from_list(lst1)
    bits2 = binned_bitsets_from_list(lst2)

    bitsets = dict()
    for key in bits1:
        if key in bits2:
            bits1[key].iand(bits2[key])
            bitsets[key] = bits1[key]

    for chrom in bitsets:
        bits = bitsets[chrom]
        end = 0
        while 1:
            start = bits.next_set(end)
            if start == bits.size:
                break
            end = bits.next_clear(start)
            yield [chrom, start, end]
    bits1.clear()
    bits2.clear()
    bitsets.clear()


def subtractBed3(lst1, lst2) -> Iterator[List]:
    """subtract lst2 from lst1"""
    bitsets1 = binned_bitsets_from_list(lst1)
    bitsets2 = binned_bitsets_from_list(lst2)

    for chrom in bitsets1:
        if chrom not in bitsets1:
            continue
        bits1 = bitsets1[chrom]
        if chrom in bitsets2:
            bits2 = bitsets2[chrom]
            bits2.invert()
            bits1.iand(bits2)
        end = 0
        while 1:
            start = bits1.next_set(end)
            if start == bits1.size:
                break
            end = bits1.next_clear(start)

            yield [chrom, start, end]

    bitsets1.clear()
    bitsets2.clear()



#!/usr/bin/env python3
"""Module enabling a sh like infix syntax (using pipes).
"""
import builtins
import functools
import gzip
import itertools
import os
import shlex
import subprocess
import sys
from collections import deque
from io import TextIOWrapper
from typing import (
    Set,
    List,
    Deque,
    Dict,
    Any,
    Tuple,
    Generic,
    Iterable,
    Iterator,
    TypeVar,
    Callable,
    Literal,
    Optional,
    Union,
    TextIO,
    overload
)

__all__ = [
    "cat",
    "zcat",
    "stdin",
    "cmdin",
    "mkdir",
    "cd",
    "Pipe",
    "head",
    "tail",
    "skip",
    "filter",
    "dedup",
    "uniq",
    "count",
    "to_list",
    "to_tuple",
    "to_dict",
    "to_set",
    "permutations",
    "join",
    "stdout",
    "write",
    "tee",
    "mapwith",
    "transform",
    "apply",
    "for_each",
    "flatten",
    "take_while",
    "drop_while",
    "skip_while",
    "kgroupby",
    "groupby",
    "groupby2",
    "sort",
    "reverse",
    "transpose",
    "window",
    "round",
    "chain",
    "concat",
    "islice"
]

_T = TypeVar("_T")
_S = TypeVar("_S")


class Pipe(Generic[_T]):
    """
    Represent a Pipeable Element :
    Described as :
    first = Pipe(lambda iterable: next(iter(iterable)))
    and used as :
    print [1, 2, 3] | first
    printing 1
    Or represent a Pipeable Function :
    It's a function returning a Pipe
    Described as :
    select = Pipe(lambda iterable, pred: (pred(x) for x in iterable))
    and used as :
    print [1, 2, 3] | select(lambda x: x * 2)
    # 2, 4, 6
    """

    def __init__(self, func: Callable[[Iterable[_T]], Any]):
        self.func = func
        functools.update_wrapper(self, func)

    def __ror__(self, other: Iterable[_T]):
        return self.func(other)

    def __call__(self, *args, **kwargs):
        return Pipe(lambda x: self.func(x, *args, **kwargs))


def cat(file: str) -> Iterator[str]:
    with open(file) as f:
        for line in f:
            line = line.strip()
            if line:
                yield line


def zcat(file: str) -> Iterator[str]:
    with gzip.open(file, "rb") as f:
        for line in f:
            line = line.decode().strip()
            if line:
                yield line


def stdin() -> Iterator[str]:
    for line in sys.stdin:
        line = line.strip()
        if line:
            yield line


def cmdin(cmd: str, shell: bool = False) -> Iterator[str]:
    if shell:
        cmd2 = cmd
    else:
        cmd2 = shlex.split(cmd)
    with subprocess.Popen(cmd2, stdout=subprocess.PIPE, encoding='utf8', shell=shell) as proc:
        if proc.stdout is None:
            raise RuntimeError(proc.stderr)
        for line in proc.stdout:
            line = line.strip()
            if line:
                yield line


def mkdir(path):
    if not os.path.exists(path):
        os.makedirs(path)


def cd(path):
    if os.path.exists(path):
        os.chdir(path)
    else:
        raise RuntimeError(f'{path} not exist')


@Pipe
def head(iterable: Iterable[_T], n: int = 10) -> Iterator[_T]:
    """Yield n of elements in the given iterable."""
    return itertools.islice(iterable, n)


@Pipe
def tail(iterable: Iterable[_T], n: int = 10) -> Deque:
    """Yield n of elements in the given iterable."""
    return deque(iterable, maxlen=n)


@Pipe
def skip(iterable: Iterable[_T], n: int) -> Iterator[_T]:
    """Skip n elements in the given iterable, then yield others."""
    return itertools.islice(iterable, n, None)


@Pipe
def filter(iterable: Iterable[_T], func: Callable[[_T], bool]) -> Iterable[_T]:
    return builtins.filter(func, iterable)


@Pipe
def dedup(iterable: Iterable[_T], key: Callable = lambda x: x) -> Iterator[_T]:
    """Only yield unique items. Use a set to keep track of duplicate data."""
    seen = set()
    for item in iterable:
        dupkey = key(item)
        if dupkey not in seen:
            seen.add(dupkey)
            yield item


@Pipe
def uniq(iterable: Iterable[_T], key: Callable = lambda x: x) -> Iterator[_T]:
    """Deduplicate consecutive duplicate values."""
    iterator = iter(iterable)
    try:
        prev = next(iterator)
    except StopIteration:
        return
    yield prev
    prevkey = key(prev)
    for item in iterator:
        itemkey = key(item)
        if itemkey != prevkey:
            yield item
        prevkey = itemkey


@overload
def count(iterable: Iterable[_T]) -> int: ...


@overload
def count(iterable: Iterable[_T], func: Callable[[int], _S]) -> _S: ...


@Pipe
def count(iterable, func=None):
    """
    Iteration will be complete
    """
    value = sum(builtins.map(lambda _: 1, iterable))
    if func is not None:
        return func(value)
    else:
        return value


@Pipe
def to_list(iterable: Iterable[_T]) -> List[_T]:
    return list(iterable)


@Pipe
def to_tuple(iterable: Iterable[_T]) -> Tuple[_T, ...]:
    return tuple(iterable)


@Pipe
def to_dict(iterable: Iterable[_T]) -> Dict[_T, _T]:
    return dict(iterable)


@Pipe
def to_set(iterable: Iterable[_T]) -> Set[_T]:
    return set(iterable)


@Pipe
def permutations(iterable: Iterable[_T], r: Optional[int] = None):
    # permutations('ABCD', 2) --> AB AC AD BA BC BD CA CB CD DA DB DC
    # permutations(range(3)) --> 012 021 102 120 201 210
    for x in itertools.permutations(iterable, r):
        yield x


@Pipe
def join(iterable: Iterable[_T], separator: str = ", "):
    return separator.join(builtins.map(str, iterable))


@Pipe
def stdout(iterable: Iterable[_T], func: Callable[[_T], Any] = str, end: str = "\n"):
    for item in iterable:
        sys.stdout.write(func(item) + end)


@Pipe
def write(iterable: Iterable[_T], path_or_buf: Union[str, TextIO], func: Callable[[_T], Any] = str, end: str = "\n"):
    if isinstance(path_or_buf, TextIOWrapper):
        ostream = path_or_buf
    else:
        ostream = open(path_or_buf, "w")
    for item in iterable:
        ostream.write(func(item) + end)
    ostream.close()


@Pipe
def tee(iterable: Iterable[_T], path_or_buf: Union[str, TextIO] = os.devnull, func: Callable[[_T], Any] = str, end: str = "\n"):
    if isinstance(path_or_buf, TextIOWrapper):
        if path_or_buf.name == "<stdout>" or path_or_buf.name == "/dev/stdout":
            ostream = open(os.devnull, "w")
        else:
            ostream = path_or_buf
    else:
        if path_or_buf == "/dev/stdout":
            ostream = open(os.devnull, "w")
        else:
            ostream = open(path_or_buf, "w")
    for item in iterable:
        ostream.write(func(item) + end)
        sys.stdout.write(func(item) + end)
    ostream.close()


@Pipe
def mapwith(iterable: Iterable[_T], func: Callable[[_T], _S]) -> Iterable[_S]:
    return builtins.map(func, iterable)


@Pipe
def transform(iterable: Iterable[_T], func: Callable[[Iterable[_T]], Iterable[_S]], *args, **kargs) -> Iterator[_S]:
    return iter(func(iterable, *args, **kargs))


@Pipe
def apply(iterable: Iterable[_T], func: Callable[[_T], Any]) -> Iterator[_T]:
    def inner(_iterable: Iterable[_T]):
        for i in _iterable:
            try:
                func(i)
                yield i
            except Exception:
                raise RuntimeError(str(i))

    return inner(iterable)


@Pipe
def for_each(iterable: Iterable[_T], func: Callable[[_T], Any]):
    """
    Iteration will be complete
    """
    for i in iterable:
        func(i)


@Pipe
def flatten(iterable: Iterable[_T], ignore_types=(str, bytes)) -> Iterator[_S]:
    for x in iterable:
        if isinstance(x, Iterable) and not isinstance(x, ignore_types):
            yield from x | flatten
        else:
            yield x


@Pipe
def take_while(iterable: Iterable[_T], func: Callable[[_T], bool]) -> Iterator[_T]:
    return itertools.takewhile(func, iterable)


@Pipe
def drop_while(iterable: Iterable[_T], func: Callable[[_T], bool]) -> Iterator[_T]:
    return itertools.dropwhile(func, iterable)


@Pipe
def skip_while(iterable: Iterable[_T], func: Callable[[_T], bool]) -> Iterator[_T]:
    for item in iterable:
        if func(item):
            continue
        yield item


@Pipe
def kgroupby(iterable: Iterable[_T], key: Optional[Callable[[_T], _S]] = None) -> Iterable[Tuple[_T, Tuple[_T, ...]]]:
    return ((i, tuple(g)) for i, g in itertools.groupby(iterable, key=key))


@Pipe
def groupby(iterable: Iterable[_T], key: Optional[Callable[[_T], _S]] = None) -> Iterable[Tuple[_T, ...]]:
    return (tuple(g) for _, g in itertools.groupby(iterable, key=key))


@Pipe
def groupby2(iterable: Iterable[_T], func: Callable[[_T, _T], bool], compare: Literal['start', 'end'] = 'start') -> Iterator[Tuple[_T, ...]]:
    """
    iterable must be sorted before
    Parameters:
    ----------
    compare: where the behind element compare to in one group

    Example:
    [1, 1, 2, 2, 3, 3, 4, 5, 4, 6, 7, 8] | groupby2(lambda x, y: y-x<2, compare='end') | stdout
    (1, 1, 2, 2, 3, 3, 4, 5, 4)
    (6, 7, 8)
    [1, 1, 2, 2, 3, 3, 4, 5, 4, 6, 7, 8] | groupby2(lambda x, y: y-x<2) | stdout
    (1, 1, 2, 2)
    (3, 3, 4)
    (5, 4, 6)
    (7, 8)
    """
    assert func is not None
    iterable = iter(iterable)
    x = deque(itertools.islice(iterable, 1))
    compared = 0 if compare == 'start' else -1
    for i in iterable:
        if func(x[compared], i):
            x.append(i)
        else:
            yield tuple(x)
            x.clear()
            x.append(i)
    if x:
        yield tuple(x)


@Pipe
def sort(iterable: Iterable[_T], key: Optional[Callable[[_T], _S]] = None, reverse: bool = False) -> Iterable[_T]:
    return sorted(iterable, key=key, reverse=reverse)


@Pipe
def reverse(iterable: Iterable[_T]) -> Iterable[_T]:
    return reversed(iterable)


@Pipe
def transpose(iterable: Iterable[_T]) -> Iterable[_T]:
    return builtins.zip(*iterable)


@Pipe
def window(iterables: Iterable[_T], window: int = 2, step: int = 1) -> Iterator[Tuple[_T, ...]]:
    assert window > 0 and step > 0
    iterable = iter(iterables)
    d = deque(itertools.islice(iterable, window), maxlen=window)
    n = 0
    yield tuple(d)
    for x in iterable:
        d.append(x)
        n += 1
        if n == step:
            yield tuple(d)
            n = 0
    if n != 0 and step > n:
        for _ in range(step-n):
            d.popleft()
        if d:
            yield tuple(d)


@Pipe
def round(iterable: Iterable[_T], func: Callable[..., bool], b: int = 1, a: int = 1) -> Iterator[_S]:
    return iterable | window(b+a+1) | filter(lambda x: func(x[b])) | flatten


chain = Pipe(itertools.chain.from_iterable)
concat = Pipe(itertools.chain)
islice = Pipe(itertools.islice)



# -*- coding: utf-8 -*-
import sys
import time
from collections import deque, defaultdict
from enum import Enum
from typing import (
    AbstractSet,
    Set,
    DefaultDict,
    Literal,
    Tuple,
    Optional,
    Iterator
)

import pysam

from pybiotk.utils import logging


class BamType(Enum):
    SE = "SE"
    PE = "PE"


def check_bam_type(filename: str) -> BamType:
    logging.info(f"checking bam type: {filename} ...")
    with pysam.AlignmentFile(filename) as bam:
        read = next(bam)
        if read.is_read1 or read.is_read2:
            bamtype = BamType.PE
        else:
            bamtype = BamType.SE
        sys.stderr.write(f"bamtype is {bamtype.value}.\n")
    return bamtype


def count_bam_size(filename: str) -> int:
    with pysam.AlignmentFile(filename) as bam:
        count = bam.count()
    return count


class BamTypeError(RuntimeError):
    pass


class Bam(pysam.AlignmentFile):
    def iter(self, flags: Optional[AbstractSet[int]] = None,
             remove_flags: Optional[AbstractSet[int]] = None,
             secondary=False, supplementary=False) -> Iterator[pysam.AlignedSegment]:
        for read in self:
            if not secondary:
                if read.is_secondary:
                    continue
            if not supplementary:
                if read.is_supplementary:
                    continue
            if flags is not None:
                if read.flag not in flags:
                    continue
            if remove_flags is not None:
                if read.flag in remove_flags:
                    continue
            yield read

    def iter_mapped(self, secondary=False, supplementary=False, qcut=None) -> Iterator[pysam.AlignedSegment]:
        for read in self:
            if read.is_unmapped:
                continue
            if read.is_qcfail:
                continue
            if read.is_duplicate:
                continue
            if not secondary:
                if read.is_secondary:
                    continue
            if not supplementary:
                if read.is_supplementary:
                    continue
            if qcut:
                if read.mapping_quality < qcut:
                    continue
            yield read

    def iter_unmapped(self, secondary=False, supplementary=False) -> Iterator[pysam.AlignedSegment]:
        for read in self:
            if not secondary:
                if read.is_secondary:
                    continue
            if not supplementary:
                if read.is_supplementary:
                    continue
            if read.is_qcfail:
                continue
            if read.is_unmapped:
                yield read

    @staticmethod
    def read_to_fastx_record(read) -> pysam.libcfaidx.FastxRecord:
        return pysam.libcfaidx.FastxRecord(
                name=read.query_name,
                comment="",
                sequence=read.get_forward_sequence(),
                quality=pysam.qualities_to_qualitystring(read.get_forward_qualities())
                )

    def to_fastx_record(self) -> Iterator[pysam.libcfaidx.FastxRecord]:
        for read in self.iter_mapped():
            yield self.read_to_fastx_record(read)

    def to_bam(self, filename: str, flags: Optional[AbstractSet[int]] = None,
               remove_flags: Optional[AbstractSet[int]] = None, header=True,
               secondary=False, supplementary=False):
        template = self if header else None
        with pysam.AlignmentFile(filename, mode="wb", template=template) as bam:
            for read in self.iter(flags=flags, remove_flags=remove_flags, secondary=secondary, supplementary=supplementary):
                bam.write(read)


class BamPE(Bam):
    def __init__(self, /, *args, **kwargs):
        super().__init__()
        self.ordered_by_name = False
        self.query_names: Set[str] = set()
        self.read1_set: DefaultDict[str, pysam.AlignedSegment] = defaultdict(type(None))
        self.read2_set: DefaultDict[str, pysam.AlignedSegment] = defaultdict(type(None))
        self.read1_unpaired_querynames: Set[str] = set()
        self.read2_unpaired_querynames: Set[str] = set()
        self.ptr: Optional[int] = None

    def is_ordered_by_name(self):
        self.ordered_by_name = True

    def to_dict(self):
        self.ptr = 0
        logging.warning("saving bam to dict, make sure you have enough memory...")
        start = time.perf_counter()
        for read in self.iter_mapped():
            self.ptr += 1
            if self.ptr % 10000 == 0:
                sys.stderr.write(f"processed {self.ptr} reads.\n")
            self.query_names.add(read.query_name)
            if read.is_read1:
                self.read1_set[read.query_name] = read
            elif read.is_read2:
                self.read2_set[read.query_name] = read
            else:
                raise BamTypeError(f"{self.filename.decode('utf-8')} seems like a single end bam.")
        end = time.perf_counter()
        logging.info(f"{len(self.query_names)} reads have been saved in {end-start:.2f}s.")

    def iter_pair(self, properly_paired=False) -> Iterator[Tuple[pysam.AlignedSegment, ...]]:
        if self.ordered_by_name:
            logging.info(f"{self.filename} is ordered by queryname, use io iter mode ...")
            d = deque(maxlen=2)
            for read in self.iter_mapped():
                d.append(read)
                if len(d) == 2:
                    if d[0].query_name == d[1].query_name:
                        if d[0].is_read1 and d[1].is_read2:
                            yield tuple(d)
                        elif d[0].is_read2 and d[1].is_read1:
                            yield d[1], d[0]
                        else:
                            if not properly_paired:
                                if d[0].is_read1 and d[1].is_read1:
                                    yield d[0], None
                                    yield d[1], None
                                elif d[0].is_read2 and d[1].is_read2:
                                    yield None, d[0]
                                    yield None, d[1]
                                else:
                                    raise BamTypeError(f"{d[0]} or {d[1]} in {self.filename} is neither read1 nor read2.")
                        d.clear()
                    else:
                        if not properly_paired:
                            if d[0].is_read1:
                                yield d[0], None
                            elif d[0].is_read2:
                                yield None, d[0]
                            else:
                                raise BamTypeError(f"{d[0]} in {self.filename} is neither read1 nor read2.")
            if d:
                if not properly_paired:
                    if d[-1].is_read1:
                        yield d[-1], None
                    elif d[-1].is_read2:
                        yield None, d[-1]
                    else:
                        raise BamTypeError(f"{d[-1]} in {self.filename} is neither read1 nor read2.")
        else:
            logging.info(f"{self.filename} is not ordered by queryname, use dict iter mode ...")
            if not self.query_names:
                self.to_dict()
            for query_name in list(self.query_names):
                read1 = self.read1_set[query_name]
                read2 = self.read2_set[query_name]

                if read1 is not None and read2 is None:
                    self.read1_unpaired_querynames.add(query_name)

                if read1 is None and read2 is not None:
                    self.read2_unpaired_querynames.add(query_name)

                if properly_paired:
                    if read1 is not None and read2 is not None:
                        yield read1, read2
                else:
                    yield read1, read2

    def iter_unpaired(self, terminal: Literal["read1", "read2"] = "read1") -> Iterator[pysam.AlignedSegment]:
        if self.read1_unpaired_querynames or self.read2_unpaired_querynames:
            if terminal == "read1":
                for query_name in list(self.read1_unpaired_querynames):
                    yield self.read1_set[query_name]
            else:
                for query_name in list(self.read2_unpaired_querynames):
                    yield self.read2_set[query_name]
        else:
            for read1, read2 in self.iter_pair(properly_paired=False):
                if terminal == "read1":
                    if read2 is None:
                        yield read1
                else:
                    if read1 is None:
                        yield read2

    def to_bam_ordered_by_name(self, filename: str):
        header = self.header.to_dict()
        header['HD']['SO'] = "queryname"
        if self.ordered_by_name:
            read1_list = []
            read2_list = []
            with pysam.AlignmentFile(filename, mode="wb", header=header) as bam:
                for read1, read2 in self.iter_pair(properly_paired=False):
                    if read1 is None:
                        read2_list.append(read2)
                    elif read2 is None:
                        read1_list.append(read1)
                    elif read1 is not None and read2 is not None:
                        bam.write(read1)
                        bam.write(read2)
                for read in read1_list:
                    bam.write(read)
                for read in read2_list:
                    bam.write(read)
        else:
            with pysam.AlignmentFile(filename, mode="wb", header=header) as bam:
                for read1, read2 in self.iter_pair(properly_paired=True):
                    bam.write(read1)
                    bam.write(read2)
                for read in self.iter_unpaired(terminal="read1"):
                    bam.write(read)
                for read in self.iter_unpaired(terminal="read2"):
                    bam.write(read)

    def to_fastx_record_pair(self) -> Iterator[Tuple[pysam.libcfaidx.FastxRecord, ...]]:
        for read1, read2 in self.iter_pair(properly_paired=True):
            yield self.read_to_fastx_record(read1), self.read_to_fastx_record(read2)

    def to_fastx_record_unpaired(self, terminal: Literal["read1", "read2"] = "read1") -> Iterator[pysam.libcfaidx.FastxRecord]:
        for read in self.iter_unpaired(terminal):
            yield self.read_to_fastx_record(read)



#!/usr/bin/env python3
import argparse
import os
import sys
import time

from pybiotk.io import GenomeFile
from pybiotk.utils import logging, ignore


def parse_args():
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument(dest="genome", type=str, help="Genome fasta file.")
    parser.add_argument("--size", dest="size", type=int, default=100000, help="dataset size.")
    parser.add_argument("--prefix", dest="prefix", type=str, default="test", help="dataset name prefix.")
    parser.add_argument("--length", dest="length", type=int, default=50, help="random sequence length.")
    parser.add_argument("--random_on", dest="random_on", nargs="+", default=None, help="random on specific chroms. default:All chroms.")
    parser.add_argument("--load2dict", dest="load", action="store_true", help="load genome into dict.")

    return parser


def main():
    parser = parse_args()
    if len(sys.argv[1:]) < 1:
        parser.parse_args(["--help"])
    args = parser.parse_args()
    faindex = args.genome + ".fai"
    regenerate = False
    if not os.path.exists(faindex):
        regenerate = True
        logging.info("Fasta index file is not exist, generating ...")
    else:
        logging.info(f"Fasta index file {faindex} found.")
    index_start = time.perf_counter()
    genome = GenomeFile(args.genome)
    index_end = time.perf_counter()
    if regenerate:
        logging.info(f"File {faindex} generated in {index_end-index_start:.2f}s.")
    genome.random_on_chroms = args.random_on

    if args.load:
        load_start = time.perf_counter()
        logging.info("Load genome into dict ...")
        genome.load_into_dict()
        load_end = time.perf_counter()
        logging.info(f"Load completed in {load_end-load_start:.2f}s.")

    fastafile = args.prefix + ".fa"
    bed_file = args.prefix + ".bed"
    size = args.size
    logging.info("Generating random sequence ...")
    start = time.perf_counter()
    with open(fastafile, "w") as fa, open(bed_file, "w", encoding="utf-8") as bed:
        for i in range(size):
            region, seq = genome.random(args.length, args.random_on)
            fa.write(f">{region}\n{seq}\n")
            bed.write(f"{region.chrom}\t{region.start}\t{region.end}\t{i+1}\t.\t{region.strand}\n")            
            width = len(str(size))
            sys.stderr.write(f"\r{i+1:{width}}/{size}\t\t\t\t{(i+1)*100/size:.2f}%")
    sys.stderr.write("\n")
    genome.close()
    end = time.perf_counter()
    logging.info(f"{size} random reads generated in {end-start:.2f}s.")
    logging.info(f"Datasets saved in {fastafile}, {bed_file}.")


@ignore
def run():
    main()


if __name__ == "__main__":
    run()



#!/usr/bin/env python3
import argparse
import os
import sys
from typing import Sequence, Literal

import pandas as pd

from pybiotk.utils import write_table, ignore


def check_for_nonnumeric(pd_series=None):
    if pd.to_numeric(pd_series, errors='coerce').isna().sum() == 0:
        return 0
    else:
        return 1


class Norm:
    def __init__(self, df: pd.DataFrame):
        self.df = df.dropna()

    def cpm(self):
        df = self.df
        # check for non-numeric values
        for i in df.columns:
            assert check_for_nonnumeric(df[i]) == 0, \
                'dataframe contains non-numeric values in {} column'.format(i)
        self.lib_size = df.sum()
        self.cpm_norm = (df * 1e6) / df.sum()

        return self.cpm_norm

    def rpkm(self, gl="Length"):
        df = self.df
        assert gl is not None, "Provide column name for gene length in bp"
        # check for non-numeric values
        for i in df.columns:
            assert check_for_nonnumeric(df[i]) == 0, \
                'dataframe contains non-numeric values in {} column'.format(i)
        self.rpkm_norm = (df.div(df[gl], axis=0) * 1e9) / df.sum()
        self.rpkm_norm = self.rpkm_norm.drop([gl], axis=1)

        return self.rpkm_norm

    def tpm(self, gl="Length"):
        df = self.df
        assert gl is not None, "Provide column name for gene length in bp"
        # check for non-numeric values
        for i in df.columns:
            assert check_for_nonnumeric(df[i]) == 0, \
                'dataframe contains non-numeric values in {} column'.format(i)
        # gene length must be in bp
        self.a = df.div(df[gl], axis=0) * 1e3
        self.tpm_norm = (self.a * 1e6) / self.a.sum()
        self.tpm_norm = self.tpm_norm.drop([gl], axis=1)

        return self.tpm_norm


def read_data(path: str, cols: Sequence[int] = (0, 5, 6)) -> pd.DataFrame:
    df = pd.read_table(path, comment="#")
    df = df.iloc[:, cols]

    if len(cols) == 3:
        samples = [os.path.basename(i).split(".")[0] for i in df.columns[2:]]
        df.columns = ["Gene", "Length"] + samples
    else:
        samples = [os.path.basename(i).split(".")[0] for i in df.columns[1:]]
        df.columns = ["Gene"] + samples
    df = df.set_index("Gene")

    return df


def main(path: str, cols: Sequence[int], method: Literal["CPM", "RPKM", "TPM"], outpath):
    df = read_data(path, cols)

    norm = Norm(df)

    if method == "CPM":
        if "Length" in df.columns:
            norm.df = norm.df.drop(['Length'], axis=1)
        data = norm.cpm()
    elif method == "RPKM":
        data = norm.rpkm()
    elif method == "TPM":
        data = norm.tpm()
    else:
        raise RuntimeError(f"{method} normalize not support!")

    data = data[data.apply(sum, axis=1) != 0]
    data = data.round(2)
    data = data.sort_values(by=data.columns[0], ascending=False)
    write_table(data, outpath, index=True)


@ignore
def run():
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("input", type=str, help="featureCount out file.")
    parser.add_argument("-c", dest="columns", type=int, nargs="+", default=[0, 6, 5], help="geneid, count, length(required by RPKM and TPM) column, 0 based.")
    parser.add_argument("-m", dest="method", type=str, default="CPM", choices=["CPM", "RPKM", "TPM"], help="normalize method, [CPM|RPKM|TPM].")
    parser.add_argument("-o", dest="output", type=argparse.FileType('w'), default=sys.stdout, help="output file.")

    args = parser.parse_args()
    main(args.input, args.columns, args.method, args.output)


if __name__ == "__main__":
    run()



#!/usr/bin/env python3
import argparse
import time
from typing import Literal, Union, TextIO, Optional

import pysam

from pybiotk.io import OpenFqGzip, Bam, BamPE, BamType, check_bam_type
from pybiotk.utils import logging


def write_recode(fileobj: Union[TextIO, OpenFqGzip], record: pysam.libcfaidx.FastxRecord, outformat: Literal['fastq', 'fasta']):
    if outformat == 'fastq':
        fileobj.write_fastx_record(record)
    else:
        fileobj.write(f">{record.name}\n{record.sequence}\n")


def main(filename: str, prefix: str, bamtype: Optional[BamType] = None, outformat: Literal['fastq', 'fasta'] = "fasta", ordered_by_name: bool = False):
    logging.info(f"start convert {filename} to {outformat} ...")
    start = time.perf_counter()
    if bamtype is None:
        bamtype = check_bam_type(filename)

    if bamtype is BamType.SE:
        if outformat == 'fasta':
            out = open(prefix + ".fasta", "w")
        else:
            out = OpenFqGzip(prefix + ".fastq.gz")
        with Bam(filename) as bam:
            logging.info(f"writing reads to {out.name} ...")
            for record in bam.to_fastx_record():
                write_recode(out, record, outformat)
        out.close()
    elif bamtype is BamType.PE:
        if outformat == 'fasta':
            out_r1 = open(prefix + ".R1.fa", "w")
            out_r2 = open(prefix + ".R2.fa", "w")
            unpaired_r1 = open(prefix + ".unpaired.R1.fa", "w")
            unpaired_r2 = open(prefix + ".unpaired.R2.fa", "w")
        else:
            out_r1 = OpenFqGzip(prefix + ".R1.fq.gz")
            out_r2 = OpenFqGzip(prefix + ".R2.fq.gz")
            unpaired_r1 = OpenFqGzip(prefix + ".unpaired.R1.fq.gz")
            unpaired_r2 = OpenFqGzip(prefix + ".unpaired.R2.fq.gz")

        with BamPE(filename) as bampe:
            bampe.ordered_by_name = ordered_by_name
            logging.info(f"writing property paired reads to {out_r1.name} {out_r2.name} ...")
            for record1, record2 in bampe.to_fastx_record_pair():
                write_recode(out_r1, record1, outformat)
                write_recode(out_r2, record2, outformat)
            logging.info(f"writing unpaired reads to {unpaired_r1.name} ...")
            for record in bampe.to_fastx_record_unpaired(terminal="read1"):
                write_recode(unpaired_r1, record, outformat)
            logging.info(f"writing unpaired reads to {unpaired_r2.name} ...")
            for record in bampe.to_fastx_record_unpaired(terminal="read2"):
                write_recode(unpaired_r2, record, outformat)
        out_r1.close()
        out_r2.close()
        unpaired_r1.close()
        unpaired_r2.close()
        end = time.perf_counter()
        logging.info(f"task finished in {end-start:.2f}s.")


def run():
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("input", type=str, help="input bam.")
    parser.add_argument("--bamtype", dest="bamtype", type=str, choices=["SE", "PE"],
                        default=None, help="input bam type single end or pair end, default:autodectect.")
    parser.add_argument("-o", dest="outprefix", type=str, required=True,
                        help="output file name prefix.")
    parser.add_argument("--outformat", dest="outformat", type=str, choices=["fasta", 'fastq'],
                        default="fastq",  help="output file format fasta or fastq, default:fastq.")
    parser.add_argument("--ordered_by_name", dest="ordered_by_name", action="store_true",
                        help="if input bam is ordered by name.")

    args = parser.parse_args()
    bamtype = None if args.bamtype is None else BamType(args.bamtype)
    main(args.input, args.outprefix, bamtype, args.outformat, args.ordered_by_name)


if __name__ == "__main__":
    run()



#!/usr/bin/env python3
import argparse
from typing import Literal

import pysam

from pybiotk.io import OpenFqGzip
from pybiotk.utils import reverse_seq


def reverse_fastx(filename: str, output: str, outformat: Literal['fastq', 'fasta'] = 'fastq'):
    if outformat == 'fastq':
        ostream = OpenFqGzip(output)
    else:
        ostream = open(output, "w")
        
    for entry in pysam.FastxFile(filename):
        name = entry.name
        sequence = reverse_seq(entry.sequence)
        comment = entry.comment
        quality = entry.quality if entry.quality is None else "".join(reversed(str(entry.quality)))
        if outformat == "fastq":
            ostream.write_entry(name, sequence, comment, quality)
        else:
            ostream.write(f">{name}\n{sequence}\n")
    ostream.close()
    

def run():
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("-i", dest="input", type=str, required=True, help="input *.fastx file.")
    parser.add_argument("-o", dest="output", type=str, required=True, help="output *.fastx file.")
    parser.add_argument("-f", dest="outformat", type=str, choices=["fasta", 'fastq'], default="fastq",  help="output file format fasta or fastq, default:fastq.")
    args = parser.parse_args()
    reverse_fastx(args.input, args.output, args.outformat)


if __name__ == "__main__":
    run()



# -*- coding: utf-8 -*-
import itertools
import random
import re
import sys
from typing import Dict, List, Sequence, Iterable, Iterator, Tuple, Literal, Callable, Optional

import pysam

from pybiotk.intervals import GRange
from pybiotk.utils import reverse_seq


class FastaFile(pysam.FastaFile):
    def __init__(self, /, *args, **kwargs):
        super().__init__()
        self.reference_dict: Optional[Dict[str, str]] = None

    def __iter__(self) -> Iterator:
        for reference in self.references:
            yield reference, self[reference]

    def stdout(self, referenceList: Optional[Sequence[str]] = None, wrap: bool = True):
        if not referenceList:
            referenceList = self.references

        for reference in referenceList:
            if reference not in self.references:
                continue
            sys.stdout.write(f">{reference}\n")
            sequence = self[reference]
            if wrap:
                seqlen = len(sequence)
                index = range(seqlen)
                for i in itertools.islice(index, None, None, 60):
                    sys.stdout.write(f"{sequence[i: i+60]}\n")
            else:
                sys.stdout.write(f"{sequence}\n")

    def load_into_dict(self) -> Dict[str, str]:
        self.reference_dict = dict((reference, seq) for reference, seq in self)
        return self.reference_dict

    def fetchs(self, reference: str, start: int, end: int, strand: Literal["+", "-"] = "+"):
        sequence = self.fetch(reference, int(start), int(end))
        if strand == "-":
            sequence = reverse_seq(sequence)
        return sequence

    @staticmethod
    def fetch_use_dict(reference_dict: Dict[str, str], reference: str, start: int, end: int, strand: Literal["+", "-"] = "+"):
        sequence = reference_dict[reference][int(start):int(end)]
        if strand == "-":
            sequence = reverse_seq(sequence)
        return sequence

    def dict_fetch(self, reference: str, start: int, end: int, strand: Literal["+", "-"] = "+"):
        assert self.reference_dict is not None
        return self.fetch_use_dict(self.reference_dict, reference, start, end, strand)


class GenomeFile(FastaFile):
    def __init__(self, /, *args, **kwargs):
        super().__init__()
        self.custom_chroms: List[str] = self.chroms_norm()

    get_chrom_length: Callable[[str], int] = FastaFile.get_reference_length
    chroms: List[str] = FastaFile.references

    def fetch_blocks(self, chrom: str, blocks: Iterable[Tuple[int, int]], strand: Literal["+", "-"] = "+") -> str:
        sequence = ""
        chromLen = self.get_chrom_length(chrom)
        for block in blocks:
            start = int(block[0]) if int(block[0]) > 0 else 0
            end = int(block[1]) if int(block[1]) < chromLen else chromLen
            if end <= start:
                continue
            sequence += self.fetch(chrom, start, end)
        seq = sequence if strand == '+' else reverse_seq(sequence)

        return seq

    def chroms_norm(self) -> List[str]:
        chroms = []
        for chrom in self.chroms:
            if chrom.startswith("chr") or re.match(r"\d.*", chrom):
                chroms.append(chrom)
        return chroms

    def random(self, length: int = 50, random_on_chroms: Optional[Sequence[str]] = None) -> Tuple[GRange, str]:
        strand = ["+", "-"][random.randint(0, 1)]
        chroms = self.custom_chroms
        if random_on_chroms:
            chroms = list(set(chroms) & set(random_on_chroms))
        chrom = chroms[random.randint(0, len(chroms)-1)]
        chrom_len = self.get_chrom_length(chrom)
        start = random.randint(0, chrom_len-length-1)
        end = start + 50
        if self.reference_dict:
            seq = self.dict_fetch(chrom, start, end, strand)
        else:
            seq = self.fetch_blocks(chrom, [(start, end)], strand).upper()
        gRange = GRange(chrom, start, end, strand)
        return gRange, seq

    def stdout(self, chromList: Optional[Sequence[str]] = None, wrap: bool = True):
        if not chromList:
            chromList = self.custom_chroms
        return super().stdout(chromList, wrap)



#!/usr/bin/env python3
import argparse
import sys

import pandas as pd


def main(inf, outf, up_out, down_out, topnumber, p_value=None, readnumber=None, fdr_value=None, inclevel=None):
    df = pd.read_table(inf)
    if p_value:
        df = df[df['PValue'] < p_value]
    if fdr_value:
        df = df[df['FDR'] < fdr_value]
    if inclevel:
        df = df[abs(df['IncLevelDifference']) > inclevel]
    if readnumber:
        df['AvgSAMPLE1'] = (df['IJC_SAMPLE_1'] + df['SJC_SAMPLE_1']) / 2
        df['AvgSAMPLE2'] = (df['IJC_SAMPLE_2'] + df['SJC_SAMPLE_2']) / 2
        df = df[(df['AvgSAMPLE1'] > readnumber) & (df['AvgSAMPLE2'] > readnumber)]
        df = df.drop('AvgSAMPLE2', axis=1)
        df = df.drop('AvgSAMPLE1', axis=1)

    df['sort_key'] = df.IncLevelDifference.abs()

    df = df.sort_values(by=['sort_key', 'FDR'], ascending=[False, True])
    df = df.drop('sort_key', axis=1)

    if topnumber:
        df = df[:topnumber]
    if outf:
        df.to_csv(outf, sep="\t", index=False)

    if up_out:
        up_df = df[df['IncLevelDifference'] > 0]
        up_df.to_csv(up_out, sep="\t", index=False)

    if down_out:
        down_df = df[df['IncLevelDifference'] < 0]
        down_df.to_csv(down_out, sep="\t", index=False)

    sys.stderr.write(f"Outfile {outf}: {df.shape[0]} genes.\n")


def run():
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("-i", "--input", dest="input_file", type=str,
                        required=True, help="Input file gnerated by rmats")
    parser.add_argument("-o", "--output", dest="output_file",
                        type=str, default=None, help="Output file")
    parser.add_argument("-o1", "--upout", dest="up_out",
                        type=str, default=None, help="Up output file")
    parser.add_argument("-o2", "--downout", dest="down_out",
                        type=str, default=None, help="Down output file") 
    parser.add_argument("-p", "--pvalue", dest="pvalue",
                        type=float, default=None, help="Pvalue cutoff")
    parser.add_argument("-f", "--fdr", dest="fdr",
                        type=float, default=None, help="FDR value cutoff")
    parser.add_argument("-l", "--inclevel", dest="inclevel",
                        type=float, default=0.01, help="IncLevel Difference cutoff")
    parser.add_argument("-r", "--readnumber", dest="readnumber",
                        type=int, default=10, help="""
                        Average junction (inclusion junction and skipping junction) reads number in either group cutoff""")
    parser.add_argument("-n", "--number", dest="topnumber",
                        type=int, default=None, help="How many records to save")

    args = parser.parse_args()

    main(args.input_file, args.output_file, args.up_out, args.down_out,
         args.topnumber, args.pvalue, args.readnumber, args.fdr, args.inclevel)


if __name__ == "__main__":
    run()



# -*- coding: utf-8 -*-
import gzip
from typing import Iterator, Tuple, Optional

import pysam


class FastqFile(pysam.FastxFile):
    def __init__(self, /,*args, **kwargs):
        super().__init__()
        self.ptr: Optional[int] = None

    def to_fasta(self) -> Iterator[str]:
        for entry in self:
            yield f">{entry.name}\n{entry.sequence}"

    def uniq(self) -> Iterator[pysam.libcfaidx.FastxRecord]:
        self.ptr = 0
        sequence_set = set()
        for entry in self:
            self.ptr += 1
            if entry.sequence not in sequence_set:
                sequence_set.add(entry.sequence)
                yield entry

    def iter_len(self) -> Iterator[int]:
        for entry in self:
            yield len(entry.sequence)


class FastqPair:
    def __init__(self, read1: str, read2: str):
        self.filename1 = read1
        self.filename2 = read2
        self.read1 = FastqFile(read1)
        self.read2 = FastqFile(read2)
        self.ptr: Optional[int] = None

    def __iter__(self) -> Iterator[Tuple[pysam.libcfaidx.FastxRecord, ...]]:
        for entry1, entry2 in zip(self.read1, self.read2):
            yield entry1, entry2

    def uniq(self) -> Iterator[Tuple[pysam.libcfaidx.FastxRecord, ...]]:
        self.ptr = 0
        unique = set()
        for entry1, entry2 in zip(self.read1, self.read2):
            if not entry1.name == entry2.name:
                raise RuntimeError(f"{entry1.name} != {entry2.name}")
            self.ptr += 1
            whole_reads = entry1.sequence + entry2.sequence
            if whole_reads not in unique:
                unique.add(whole_reads)
                yield entry1, entry2

    def close(self):
        self.read1.close()
        self.read2.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, traceback):
        self.close()


class OpenFqGzip:
    def __init__(self, filename, mode="wb", compresslevel=9, encoding=None, errors=None, newline=None):
        self.filename = filename
        self.mode = mode
        self.compresslevel = compresslevel
        self.encoding = encoding
        self.errors = errors
        self.newline = newline
        self.fq = gzip.open(filename, mode, compresslevel, encoding, errors, newline)
        self.name = self.filename

    def write_entry(self, name: str, sequence: str, comment: Optional[str] = None, quality: Optional[str] = None):
        if quality is None:
            quality = "F"*len(sequence)
        if comment is None:
            self.fq.write(f"@{name}\n{sequence}\n+\n{quality}\n".encode("utf-8"))
        else:
            self.fq.write(f"@{name} {comment}\n{sequence}\n+\n{quality}\n".encode("utf-8"))

    def write_fastx_record(self, fq: pysam.libcfaidx.FastxRecord):
        self.write_entry(fq.name, fq.comment, fq.sequence, fq.quality)

    def write(self, string: str):
        self.fq.write(string)

    def close(self):
        self.fq.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, traceback):
        self.fq.close()



# -*- coding: utf-8 -*-
from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, List, Dict, Literal, Tuple, Union, Optional

from pybiotk.bx.intersection import IntervalTree

from pybiotk.intervals.merge_intervals import MergedIntervals
from pybiotk.utils import is_overlap


@dataclass
class GRange:
    chrom: str
    start: int
    end: int
    strand: Optional[Literal['+', '-']] = field(default=None)
    value: List[Any] = field(default_factory=list, init=False, repr=True, compare=False)

    def __len__(self):
        return self.end - self.start

    def __str__(self) -> str:
        return f"{self.chrom}:{self.start}-{self.end}({self.strand})"

    def is_overlap(self, other: GRange) -> bool:
        if self.chrom != other.chrom:
            return False
        if self.strand != other.strand:
            return False
        if not is_overlap((self.start, self.end), (other.start, other.end)):
            return False
        return True


class GRangeTree:
    def __init__(self, strand: bool = False):
        self.strand: bool = strand
        self._trees: Dict[Union[str, Tuple[str, ...]], IntervalTree] = {}

    def __iter__(self):
        for _, tree in sorted(self._trees.items()):
            nodes = list()
            fn = nodes.append
            tree.traverse(fn)
            for node in nodes:
                yield node.interval

    def _add_record(self, key: Union[str, Tuple[str, ...]], start: int, end: int, record: Any):
        if key not in self._trees.keys():
            self._trees[key] = IntervalTree()
        self._trees[key].insert(int(start), int(end), record)

    def add(self, obj: Any, chrom: str, start: int, end: int, strand: Optional[Literal['+', '-']] = None):
        if self.strand:
            key = (chrom, strand)
        else:
            key = chrom
        self._add_record(key, start, end, obj)

    def add_range(self, record: GRange):
        if self.strand:
            key = (record.chrom, record.strand)
        else:
            key = record.chrom
        start = record.start
        end = record.end
        self._add_record(key, start, end, record)

    def find(self, chrom: str, start: int, end: int, strand: Optional[Literal['+', '-']] = None):
        if self.strand:
            key = (chrom, strand)
        else:
            key = chrom
        overlapped = []
        if key in self._trees.keys():
            overlapped = self._trees[key].find(int(start), int(end))
        return overlapped

    def iter_chrom(self, chrom: str, strand: Optional[Literal['+', '-']] = None):
        if self.strand:
            key = (chrom, strand)
        else:
            key = chrom
        nodes = []
        tree = self._trees[key]
        fn = nodes.append
        tree.traverse(fn)
        for node in nodes:
            yield node.interval


class MergedGRange:
    def __init__(self, strand: bool = False):
        self.strand: bool = strand
        self._trees: Dict[Union[str, Tuple[str, ...]], MergedIntervals] = {}

    def __iter__(self):
        for key in self._trees:
            merge_intervals = self._trees[key]
            if self.strand:
                chrom, strand = key
            else:
                chrom = key
                strand = None
            for start, end in merge_intervals:
                yield GRange(chrom, start, end, strand)

    def _add_record(self, key: Union[str, Tuple[str, ...]], start: int, end: int):
        if key not in self._trees.keys():
            self._trees[key] = MergedIntervals()
        self._trees[key].add_interval((int(start), int(end)))

    def add(self, chrom: str, start: int, end: int, strand: Optional[Literal['+', '-']] = None):
        if self.strand:
            key = (chrom, strand)
        else:
            key = chrom
        self._add_record(key, start, end)

    def add_range(self, record: GRange):
        if self.strand:
            key = (record.chrom, record.strand)
        else:
            key = record.chrom
        start = record.start
        end = record.end
        self._add_record(key, start, end)

    def intersect(self, other: MergedGRange):
        assert self.strand == other.strand
        for key in self._trees:
            if key in other._trees:
                self._trees[key].intersect(other._trees[key])

    def subtract(self, other: MergedGRange):
        assert self.strand == other.strand
        for key in self._trees:
            if key in other._trees:
                self._trees[key].subtract(other._trees[key])



"""
Joins two paired-end reads on the overlapping ends. need fastq-join 1.3.1.
"""
import argparse
# !/usr/bin/env python3
import os
import subprocess
import sys
from typing import Literal

from pybiotk.io import FastqPair, OpenFqGzip
from pybiotk.utils import logging, reverse_seq
from pybiotk.utils.reverse_fastx import reverse_fastx
from stream import mkdir


def fastq_join(fq1: str, fq2: str, outprefix: str, threads: int = 1, 
               save_as: Literal["read1", "read2"] = "read1", collapse: bool = False):
    fq1_uncom = fq1 + ".uncom.fq"
    fq2_uncom = fq2 + ".uncom.fq"
    outdir = os.path.dirname(outprefix)
    if outdir:
        mkdir(outdir)
    logging.info("start decompressing fastq ...")
    try:
        subprocess.check_call(f"pigz -p {threads} -d -c {fq1} > {fq1_uncom}", shell=True)
        subprocess.check_call(f"pigz -p {threads} -d -c {fq2} > {fq2_uncom}", shell=True)
    except subprocess.CalledProcessError:
        logging.warning(f"pigz is not installed, use gzip instead.")
        subprocess.check_call(f"gzip -d -c {fq1} > {fq1_uncom}", shell=True)
        subprocess.check_call(f"gzip -d -c {fq2} > {fq2_uncom}", shell=True)
    
    un1_uncom = outprefix + ".unmerge_R1.fq"
    un2_uncom = outprefix + ".unmerge_R2.fq"
    join_uncom = outprefix + ".merge.fq"

    logging.info("start join fastq ...")
    try:
        subprocess.check_call(f"fastq-join {fq1_uncom} {fq2_uncom} -o {un1_uncom} -o {un2_uncom} -o {join_uncom}",
                              shell=True, stdout=sys.stderr, stderr=sys.stderr)
    except subprocess.CalledProcessError:
        logging.error(f"fastq-join is not installed.")
        raise
    if os.path.exists(join_uncom):
        logging.info("join completed, remove decompressed fastq")
        os.remove(fq1_uncom)
        os.remove(fq2_uncom)
    
    logging.info("start compressing fastq ...")
    try:
        subprocess.check_call(f"pigz -f -p {threads} {un1_uncom}", shell=True)
        subprocess.check_call(f"pigz -f -p {threads} {un2_uncom}", shell=True)
        subprocess.check_call(f"pigz -f -p {threads} {join_uncom}", shell=True)
    except subprocess.CalledProcessError:
        logging.warning(f"pigz is not installed, use gzip instead.")
        subprocess.check_call(f"gzip -f {un1_uncom}", shell=True)
        subprocess.check_call(f"gzip -f {un2_uncom}", shell=True)
        subprocess.check_call(f"gzip -f {join_uncom}", shell=True)
    logging.info("compressing completed.")
    
    un1 = un1_uncom + ".gz"
    un2 = un2_uncom + ".gz"
    join = join_uncom + ".gz"
    
    if save_as == "read2":
        logging.info(f"start to reverse sequence ...")
        rev_join = join_uncom + ".rev.gz"
        reverse_fastx(join, rev_join)
        os.remove(join)
        os.rename(rev_join, join)
    
    if collapse:
        un_collapse = outprefix + ".unmerge.fq.gz"
        with OpenFqGzip(un_collapse) as fq, FastqPair(un1, un2) as fp:
            for entry1, entry2 in fp:
                name1 = entry1.name + "_R1"
                sequence1 = entry1.sequence
                comment1 = entry1.comment
                quality1 = entry1.quality
                name2 = entry2.name + "_R2"
                sequence2 = entry2.sequence
                comment2 = entry2.comment
                quality2 = entry2.quality
                
                if save_as == "read2":
                    sequence1 = reverse_seq(sequence1)
                    quality1 = "".join(reversed(quality1))
                else:
                    sequence2 = reverse_seq(sequence2)
                    quality2 = "".join(reversed(quality2))
                fq.write_entry(name1, sequence1, comment1, quality1)
                fq.write_entry(name2, sequence2, comment2, quality2)
        merge_collapse = outprefix + ".collapse.fq.gz"
        subprocess.check_call(f"cat {join} {un_collapse} > {merge_collapse}", shell=True)
        if os.path.exists(merge_collapse):
            logging.info(f"collapse completed, remove temp fastq files")
            os.remove(un1)
            os.remove(un2)
            os.remove(join)
            os.remove(un_collapse)


def run():
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("-1", dest="fq1", type=str, required=True, help="R1 fastq.")
    parser.add_argument("-2", dest="fq2", type=str, required=True, help="R2 fastq.")
    parser.add_argument("-o", dest="outprefix", type=str, default="fastq", help="output file prefix.")
    parser.add_argument("-p", dest="threads", type=int, default=1, help="use pgzip")
    parser.add_argument("--save_as", dest="save_as", default="read1", choices=["read1", "read2"], help="save as read1 or read2.")
    parser.add_argument("--collapse", dest="collapse", action="store_true", help="collapse merged and unmerged fastq.")
    
    args = parser.parse_args()
    fastq_join(args.fq1, args.fq2, args.outprefix, args.threads, args.save_as, args.collapse)


if __name__ == "__main__":
    run()



# -*- coding: utf-8 -*-
from __future__ import annotations

import warnings
from typing import List, Tuple, Literal, Iterable, Optional, TYPE_CHECKING

from pybiotk.annodb import GFeature, GenomicAnnotation
from stream import flatten, window, skip_while, to_list

if TYPE_CHECKING:
    from pybiotk.io import TransInfo, Bed12, GTF


class Transcript(GFeature):
    def __init__(
        self,
        transcript_id: Optional[str] = None,
        transcript_name: Optional[str] = None,
        transcript_type: Optional[str] = None,
        gene_id: Optional[str] = None,
        gene_name: Optional[str] = None,
        gene_type: Optional[str] = None,
        chrom: Optional[str] = None,
        start: int = 0,
        end: int = 0,
        strand: Optional[Literal['+', '-']] = None,
        cds_start: Optional[int] = None,
        cds_end: Optional[int] = None,
        exons: Iterable[Tuple[int, int]] = ()
    ):
        self.transcript_id = transcript_id
        self.transcript_name = transcript_name
        self.transcript_type = transcript_type
        self.gene_id = gene_id
        self.gene_name = gene_name
        self.gene_type = gene_type
        self.chrom = chrom
        self.start = int(start)
        self.end = int(end)
        self.strand = strand
        self.cds_start = int(cds_start) if cds_start is not None else cds_start
        self.cds_end = int(cds_end) if cds_end is not None else cds_end
        self._exons = list(exons)
        self._introns = None
        self._cds_exons = None
        self._utr5_exons = None
        self._utr3_exons = None

    def __repr__(self) -> str:
        return f"{self.transcript_id}:{self.chrom}:{self.start}-{self.end}({self.strand})"

    __str__ = __repr__

    @classmethod
    def init_by_gtf(cls, gtf: GTF):
        return cls(gtf.transcript_id(), gtf.transcript_name(), gtf.transcript_type(),
                   gtf.gene_id(), gtf.gene_name(), gtf.gene_type(), gtf.chrom,
                   gtf.start - 1, gtf.end, gtf.strand, 0, 0, [(gtf.start-1, gtf.end)])

    def update(self, gtf: GTF):
        self.end = gtf.end
        self._exons.append((gtf.start-1, gtf.end))

    @classmethod
    def init_by_bed(cls, transinfo: TransInfo, bed12: Bed12):
        return cls(*transinfo.__slots__, bed12.chrom, bed12.start, bed12.end,
                   bed12.strand, bed12.thickStart, bed12.thickEnd, bed12.exons())

    def is_protein_coding(self) -> bool:
        if self.transcript_type == 'protein_coding':
            return True
        else:
            return False

    def exons(self) -> List[Tuple[int, int]]:
        return self._exons

    def introns(self) -> List[Tuple[int, int]]:
        if self._introns is None:
            if len(self._exons) <= 1:
                self._introns = []
            else:
                self._introns = [self.start, *(self.exons() | flatten), self.end] | window(2, 2) | skip_while(lambda x: x[0] == x[1]) | to_list
        return self._introns

    def tss_region(self, region: Tuple[int, int] = (-1000, 1000)) -> Tuple[int, ...]:
        if self.strand == '+':
            region = tuple(i+self.start for i in region)
        else:
            region = tuple(reversed(tuple(self.end-i for i in region)))
        return region

    def downstream(self, down: int = 5000) -> Tuple[int, int]:
        if self.strand == '+':
            downstream = (self.end, self.end+down)
        else:
            downstream = (self.start-down, self.start)
        return downstream

    def _classify_exons(self):
        if self.cds_start is not None and self.cds_end is not None:
            utr5_exons = []
            utr3_exons = []
            cds_exons = []
            for exon in self._exons:
                if exon[1] <= self.cds_start:
                    utr5_exons.append(exon)
                elif exon[0] >= self.cds_end:
                    utr3_exons.append(exon)
                elif self.cds_start <= exon[0] and exon[1] <= self.cds_end:
                    cds_exons.append(exon)
                elif exon[0] < self.cds_start < exon[1] <= self.cds_end:
                    utr5_exons.append((exon[0], self.cds_start))
                    cds_exons.append((self.cds_start, exon[1]))
                elif self.cds_start <= exon[0] < self.cds_end < exon[1]:
                    cds_exons.append((exon[0], self.cds_end))
                    utr3_exons.append((self.cds_end, exon[1]))
                elif exon[0] < self.cds_start and self.cds_end < exon[1]:
                    utr5_exons.append((exon[0], self.cds_start))
                    cds_exons.append((self.cds_start, self.cds_end))
                    utr3_exons.append((self.cds_end, exon[1]))

            if self.strand == '-':
                utr5_exons, utr3_exons = utr3_exons, utr5_exons
            loginfo = f"exons: {self.exons()}\ncds_start: {self.cds_start}, cds_end: {self.cds_end}"
            if not cds_exons:
                warnings.warn(f"utr5_exons of {self.transcript_id} does not exist, please check:\n{loginfo}")
            self._utr5_exons = utr5_exons
            self._cds_exons = cds_exons
            self._utr3_exons = utr3_exons
        else:
            self._utr5_exons = []
            self._cds_exons = []
            self._utr3_exons = []

    def cds_exons(self) -> List[Tuple[int, int]]:
        if self._cds_exons is None:
            self._classify_exons()
        return self._cds_exons

    def utr5_exons(self) -> List[Tuple[int, int]]:
        if self._utr5_exons is None:
            self._classify_exons()
        return self._utr5_exons

    def utr3_exons(self) -> List[Tuple[int, int]]:
        if self._utr3_exons is None:
            self._classify_exons()
        return self._utr3_exons

    def annotation(self, blocks: List[Tuple[int, int]], tss_region: Tuple[int, int] = (-1000, 1000), downstream: int = 3000) -> GenomicAnnotation:
        return GenomicAnnotation(self.transcript_id, self.gene_name, self.start, self.end, self.transcript_type, self.anno(blocks, tss_region, downstream))



# -*- coding: utf-8 -*-
import importlib
import importlib.resources
import logging
import os
import subprocess
import sys
import warnings
from functools import wraps
from io import TextIOWrapper
from types import ModuleType
from typing import List, Dict, Sequence, Tuple, Literal, Iterator, Iterable, Optional, Callable, Union, TextIO

import pandas as pd

logging.basicConfig(format="%(asctime)s %(levelname)s: %(message)s", datefmt='%Y-%m-%d %A %H:%M:%S', level=logging.INFO)


def reverse_seq(seq: str) -> str:
    seq = seq.upper()
    complement = {
        "A": "T", "C": "G", "G": "C", "T": "A",
        "N": "N", "W": "W", "S": "S", "K": "M",
        "M": "K", "Y": "R", "R": "Y", "H": "D",
        "B": "V", "V": "B", "D": "H", "&": "&",
    }
    return "".join(reversed([complement[i] for i in seq]))


def is_overlap(interval1: Tuple[int, int], interval2: Tuple[int, int]) -> bool:
    x1, y1 = interval1
    x2, y2 = interval2
    if y1 <= x2 or y2 <= x1:
        return False
    else:
        return True


def intervals_is_overlap(intervals1: List[Tuple[int, int]], intervals2: List[Tuple[int, int]]) -> bool:
    if not is_overlap((intervals1[0][0], intervals1[-1][1]), (intervals2[0][0], intervals2[-1][1])):
        return False
    for interval1 in intervals1:
        for interval2 in intervals2:
            if is_overlap(interval1, interval2):
                return True
    return False


def cigartuples2blocks(start: int, cigartuples: Sequence[Tuple[int, int]], shift: int = 0) -> List[Tuple[int, int]]:
    tmp_abs_start = shift + start
    blocks = []
    for cigar, length in cigartuples:
        if cigar in [0, 2]:
            blocks.append((tmp_abs_start, tmp_abs_start + length))
            tmp_abs_start += length
        elif cigar in [3, 6, 10]:
            tmp_abs_start += length
    merged_blocks = []
    tmp_start = None
    block_num = len(blocks)
    for i in range(block_num):
        if tmp_start is None:
            tmp_start = blocks[i][0]
        if not ((i < block_num - 1) and (blocks[i + 1][0] <= blocks[i][1])):
            merged_blocks.append((tmp_start, blocks[i][1]))
            tmp_start = None
    return merged_blocks


def split_discontinuous(discontinuous_list: Iterable[int]) -> Iterator[List[int]]:
    continuous_list = []
    for p in discontinuous_list:
        if not continuous_list:
            continuous_list.append(p)
        elif p - continuous_list[-1] == 1:
            continuous_list.append(p)
        else:
            yield continuous_list
            continuous_list = [p]
    if continuous_list:
        yield continuous_list


def read_table(filepath_or_buffer: Union[str, TextIO], header: Optional[int] = 0, sep: Optional[str] = None, **kwargs) -> pd.DataFrame:
    if isinstance(filepath_or_buffer, TextIOWrapper):
        filename = filepath_or_buffer.name
    else:
        filename = filepath_or_buffer
    ext = os.path.splitext(filename)[1]
    try:
        if ext == "csv":
            deli = sep if sep else ","
            df = pd.read_csv(filepath_or_buffer, delimiter=deli, header=header, **kwargs)
        elif ext == "xlsx":
            df = pd.read_excel(filepath_or_buffer, header=header, **kwargs)
        else:
            deli = sep if sep else "\t"
            df = pd.read_table(filepath_or_buffer, delimiter=deli, header=header, **kwargs)
    except pd.errors.EmptyDataError:
        df = pd.DataFrame()
    if isinstance(filepath_or_buffer, TextIOWrapper):
        filepath_or_buffer.close()
    return df


def write_table(out_df: pd.DataFrame, path_or_buf: Union[str, TextIO], index: bool = False, header: bool = True, sep: str = "\t", **kwargs):
    if isinstance(path_or_buf, TextIOWrapper):
        filename = path_or_buf.name
    else:
        filename = path_or_buf
    ext = os.path.splitext(filename)[1]
    if ext == "csv":
        out_df.to_csv(path_or_buf, index=index, header=header, sep=",", **kwargs)
    elif ext == "xlsx":
        out_df.to_excel(path_or_buf, index=index, header=header, **kwargs)
    else:
        out_df.to_csv(path_or_buf, index=index, header=header, sep=sep, **kwargs)

    if isinstance(path_or_buf, TextIOWrapper):
        path_or_buf.close()


def blocks_len(blocks: Iterable[Tuple[int, int]]) -> int:
    return sum(y-x for (x, y) in blocks)


def infer_fragment_strand(strand: Literal["+", "-"], rule: str = "1+-,1-+,2++,2--", is_read2: bool = False):
    fragment_strand = None
    rules = ["1+-,1-+,2++,2--", "1++,1--,2+-,2-+", "+-,-+", "++,--"]
    assert rule in rules
    if rule == rules[0] or rule == rules[2]:
        if strand == "-":
            fragment_strand = "+"
        elif strand == "+":
            fragment_strand = "-"
    else:
        if strand == "+":
            fragment_strand = "+"
        elif strand == "-":
            fragment_strand = "-"

    if is_read2:
        if fragment_strand == "+":
            fragment_strand = "-"
        else:
            fragment_strand = "+"
    return fragment_strand


def ignore(func: Callable):
    @wraps(func)
    def wrapper(*args, **kargs):
        try:
            traceback = func(*args, **kargs)
        except BrokenPipeError:
            sys.stdout = None
            sys.exit(0)
        except KeyboardInterrupt:
            sys.stdout = None
            sys.exit(0)
        return traceback
    return wrapper


def load_chrom_length_dict(filepath_or_buffer: Union[str, TextIOWrapper]) -> Dict[str, int]:
    if isinstance(filepath_or_buffer, TextIOWrapper):
        buffer = filepath_or_buffer
    else:
        buffer = open(filepath_or_buffer)
    chrom_length_dict = {}
    with buffer as f:
        for line in f:
            fileds = line.split()
            chrom_length_dict[fileds[0]] = int(fileds[1])

    return chrom_length_dict


def package_resource(package: Union[str, ModuleType], resource: Union[str, os.PathLike]):
    with importlib.resources.path(package, resource) as path:
        pathstr = str(path)
    return pathstr


def default_hg38_chrom_length_dict() -> Dict[str, int]:
    return load_chrom_length_dict(importlib.resources.open_text("pybiotk.data", "hg38.chrom.sizes"))


def default_mm10_chrom_length_dict() -> Dict[str, int]:
    return load_chrom_length_dict(importlib.resources.open_text("pybiotk.data", "mm10.chrom.sizes"))


def bedtools_sort(path: str):
    logging.info(f"start to sort {path}...")
    tmp_path = path + ".tmp"
    os.rename(path, tmp_path)
    cmd = f"bedtools sort -i {tmp_path} > {path}"
    try:
        subprocess.check_call(cmd, shell=True)
        os.remove(tmp_path)
    except subprocess.CalledProcessError:
        os.rename(tmp_path, path)
        warnings.warn(f"{path} is unsorted, maybe bedtools is not installed.")



#!/usr/bin/env python3
import argparse
import os
import sys
import time
from typing import Literal, Tuple, TextIO

from pybiotk.annodb import AnnoSet, Gene
from pybiotk.intervals import GRangeTree, merge_intervals
from pybiotk.io import (
    GtfFile, Openbed, BamType, Bam, BamPE, check_bam_type
)
from pybiotk.utils import logging, infer_fragment_strand, intervals_is_overlap
from stream import groupby


def load_grangetree(
    filename: str,
    level: Literal["transcript", "gene"] = "transcript",
    tss_region: Tuple[int, int] = (-1000, 1000),
    downstream: int = 3000,
    strand: bool = True
) -> GRangeTree:

    start = time.perf_counter()
    logging.info(f"start to load {level} grangetree ...")
    grangetree = GRangeTree(strand)
    with GtfFile(filename) as gtf:
        i = 0
        if level == "transcript":
            for transcript in gtf.to_transcript():
                grange = [*transcript.tss_region(tss_region), *transcript.downstream(downstream)]
                grange.sort()
                grangetree.add(transcript, transcript.chrom, grange[0], grange[-1], transcript.strand)
                del grange
                i += 1
                if i % 1000 == 0:
                    sys.stderr.write(f"\rload {i:6} {level}s.")
        else:
            for transcripts in gtf.to_transcript() | groupby(lambda x: x.gene_id):
                gene = Gene.init_by_transcipts(transcripts)
                grange = [*gene.tss_region(tss_region), *gene.downstream(downstream)]
                grange.sort()
                grangetree.add(gene, gene.chrom, grange[0], grange[-1], gene.strand)
                del grange
                i += 1
                if i % 1000 == 0:
                    sys.stderr.write(f"\rload {i:6} {level}s.")
        sys.stderr.write(f"\rload {i:6} {level}s.\n")
    end = time.perf_counter()
    logging.info(f"grangetree load completed in {end-start}s.")
    return grangetree


def annobam(filename: str,
            file_obj: TextIO,
            grangetree: GRangeTree,
            anno_fragments: bool = False,
            tss_region: Tuple[int, int] = (-1000, 1000),
            downstream: int = 3000,
            rule: str = "1+-,1-+,2++,2--",
            ordered_by_name: bool = False
            ):
    bamtype = check_bam_type(filename)

    def anno_read(_blocks, _strand, _read):
        start = int(_blocks[0][0])
        end = int(_blocks[-1][1])
        fragment_strand = infer_fragment_strand(_strand, rule, _read.is_read2)
        genes = grangetree.find(_read.reference_name, start, end, fragment_strand)
        if not genes:
            file_obj.write(f"{_read.query_name}\t{_read.reference_name}\t{start}\t{end}\t{_blocks}\t{fragment_strand}\tIntergenic\t*\t*\t*\t*\t*\n")
        else:
            annoset = AnnoSet(gene.annotation(_blocks, tss_region, downstream) for gene in genes)
            file_obj.write(f"{_read.query_name}\t{_read.reference_name}\t{start}\t{end}\t{_blocks}\t{fragment_strand}\t{annoset}\n")

    if bamtype is BamType.PE and anno_fragments:
        with BamPE(filename) as bam:
            bam.ordered_by_name = ordered_by_name
            for read1, read2 in bam.iter_pair(properly_paired=False):
                if read1 is not None and read2 is not None:
                    strand1 = "-" if read1.is_reverse else "+"
                    strand2 = "-" if read2.is_reverse else "+"
                    blocks1 = read1.get_blocks()
                    blocks2 = read2.get_blocks()
                    if read1.reference_name == read2.reference_name and strand1 == strand2 and intervals_is_overlap(blocks1, blocks2):
                        merge_blocks = merge_intervals(blocks1 + blocks2)
                        anno_read(merge_blocks, strand1, read1)
                    else:
                        anno_read(blocks1, strand1, read1)
                        anno_read(blocks2, strand2, read2)
                if read1 is not None and read2 is None:
                    blocks = read1.get_blocks()
                    strand = '-' if read1.is_reverse else '+'
                    anno_read(blocks, strand, read1)
                if read2 is not None and read1 is None:
                    blocks = read2.get_blocks()
                    strand = '-' if read2.is_reverse else '+'
                    anno_read(blocks, strand, read2)
    else:
        with Bam(filename) as bam:
            i = 0
            for read in bam.iter_mapped():
                blocks = read.get_blocks()
                strand = '-' if read.is_reverse else '+'
                anno_read(blocks, strand, read)
                i += 1
            sys.stderr.write(f"annotate {i} reads.\n")


def annobed(filename: str,
            file_obj: TextIO,
            grangetree: GRangeTree,
            tss_region: Tuple[int, int] = (-1000, 1000),
            downstream: int = 3000
            ):
    with Openbed(filename) as bedfile:
        i = 0
        for bed in bedfile:
            genes = grangetree.find(bed.chrom, bed.start, bed.end, bed.strand)
            if not genes:
                file_obj.write(f"{bed.name}\t{bed.chrom}\t{bed.start}\t{bed.end}\t{bed.strand}\tIntergenic\t*\t*\t*\t*\t*\n")
            else:
                annoset = AnnoSet(gene.annotation([(bed.start, bed.end)], tss_region, downstream) for gene in genes)
                file_obj.write(f"{bed.name}\t{bed.chrom}\t{bed.start}\t{bed.end}\t{bed.strand}\t{annoset}\n")
            i += 1
        sys.stderr.write(f"annotate {i} beds.\n")


def main(
    filename: str,
    outfilename: str,
    gtf_file: str,
    level: Literal["transcript", "gene"] = "transcript",
    tss_region: Tuple[int, int] = (-1000, 1000),
    downstream: int = 3000,
    strand: bool = True,
    rule: str = "1+-,1-+,2++,2--",
    annofragments: bool = False,
    ordered_by_name: bool = False
):
    start = time.perf_counter()
    filetype = os.path.splitext(filename)[1]
    grangetree = load_grangetree(gtf_file, level, tss_region, downstream, strand)
    with open(outfilename, "w", encoding="utf-8") as annofile:
        if filetype == ".bam":
            annofile.write("seqname\tchrom\tstart\tend\tblocks\tstrand\tannotation\tgeneStart\tgeneEnd\tgeneName\tid\ttype\n")
            logging.info("start annotating, use bam mode ...")
            annobam(filename, annofile, grangetree, annofragments, tss_region, downstream, rule, ordered_by_name)
        elif filetype.startswith(".bed"):
            annofile.write("seqname\tchrom\tstart\tend\tstrand\tannotation\tgeneStart\tgeneEnd\tgeneName\tid\ttype\n")
            logging.info("start annotating, use bed mode ...")
            annobed(filename, annofile, grangetree, tss_region, downstream)
        else:
            raise RuntimeError(f"Unable to infer file type as bam or bed from filename: {filename}.")
    end = time.perf_counter()
    logging.info(f"task completed in {end-start:.2f}s, annofile saved in {outfilename}")


def run():
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("-i", "--input", dest="input", type=str, required=True,
                        help="input file, bam or bed. The file type will be inferred from the filename suffix ['*.bam', '*.bed'].")
    parser.add_argument("-o", "--output", dest="output", type=str, required=True, help="output file name.")
    parser.add_argument("-g", "--gtf", dest='gtf', required=True,
                        help="gtf file download from Genecode, or a sorted gtf file.")
    parser.add_argument("-l", "--level", dest="level", type=str, default="transcript", choices=["transcript", "gene"],
                        help="annotation level, transcript or gene.")
    parser.add_argument("--tss_region", dest="tss_region", type=int, nargs="+", default=[-1000, 1000],
                        help="choose region from tss.")
    parser.add_argument("--downstream", dest="downstream", type=int, default=3000, help="downstream length from tes.")
    parser.add_argument("-s", "--strand", dest="strand", action="store_true", help="require same strandedness.")
    parser.add_argument("--rule", dest="rule", type=str, default="1+-,1-+,2++,2--",
                        choices=["1+-,1-+,2++,2--", "1++,1--,2+-,2-+", "+-,-+", "++,--"],
                        help="how read(s) were stranded during sequencing. only for bam.")
    parser.add_argument("-p", "--pair", dest="pair", action="store_true",
                        help="annotate fragments instead of reads.")
    parser.add_argument("--ordered_by_name", dest="ordered_by_name", action="store_true",
                        help="if input bam is ordered by name, only for pair-end bam.")

    args = parser.parse_args()

    assert len(args.tss_region) == 2, "tss_region must be a tuple of 2 elements."
    main(args.input, args.output, args.gtf, args.level, args.tss_region, args.downstream, args.strand, args.rule, args.pair, args.ordered_by_name)


if __name__ == "__main__":
    run()



#!/usr/bin/env python3
import argparse

from pybiotk.io import BamPE, BamType, BamTypeError, check_bam_type
from pybiotk.utils import logging


def main(filename: str, output: str, ordered_by_name: bool = False):
    bamtype = check_bam_type(filename)

    if bamtype is not BamType.PE:
        raise BamTypeError(f"{filename} is not a pair end bam.")

    with BamPE(filename) as bam:
        bam.ordered_by_name = ordered_by_name
        bam.to_bam_ordered_by_name(output)

    logging.info(f"output bam have been saved in {output}.")


def run():
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("input", type=str, help="input pair end bam.")
    parser.add_argument("-o", dest="output", type=str, required=True,
                        help="output file.")
    parser.add_argument("--ordered_by_name", dest="ordered_by_name", action="store_true",
                        help="if input bam is ordered by name.")

    args = parser.parse_args()
    main(args.input, args.output, args.ordered_by_name)


if __name__ == "__main__":
    run()



# -*- coding: utf-8 -*-
from __future__ import annotations

import itertools
import os
from collections import deque
from functools import partial
from typing import List, Dict, Tuple, Literal, Iterable, Iterator, Sequence, Optional, TYPE_CHECKING

from pybiotk.intervals import merge_intervals
from pybiotk.io.bed import Bed6
from pybiotk.utils import bedtools_sort
from stream import window, to_list, filter, transform, mapwith, apply, uniq, flatten

if TYPE_CHECKING:
    from pybiotk.io import GtfFile
    from pybiotk.annodb import Transcript


class MergedTranscript:
    def __init__(
        self,
        transcript_id: Optional[str] = None,
        transcript_name: Optional[str] = None,
        transcript_type: Optional[str] = None,
        gene_id: Optional[str] = None,
        gene_name: Optional[str] = None,
        gene_type: Optional[str] = None,
        chrom: Optional[str] = None,
        start: int = 0,
        end: int = 0,
        strand: Optional[Literal['+', '-']] = None,
        cds_start: Optional[int] = None,
        cds_end: Optional[int] = None,
        exons: Iterable[Tuple[int, int]] = (),
        count: int = 0,
        before: Optional[int] = None,
        after: Optional[int] = None,
    ):
        self.transcript_id = transcript_id
        self.transcript_name = transcript_name
        self.transcript_type = transcript_type
        self.gene_id = gene_id
        self.gene_name = gene_name
        self.gene_type = gene_type
        self.chrom = chrom
        self.start = int(start)
        self.end = int(end)
        self.strand = strand
        self.cds_start = int(cds_start) if cds_start is not None else cds_start
        self.cds_end = int(cds_end) if cds_end is not None else cds_end
        self.exons = exons
        self.count = count
        self.before = int(before) if before is not None else before
        self.after = int(after) if after is not None else after

    def __repr__(self) -> str:
        return f"{self.gene_name}:{self.chrom}:{self.start}-{self.end}({self.strand})"

    __str__ = __repr__

    @classmethod
    def init_by_transcripts(cls, transcripts: Iterable[Transcript]):
        min_start = float("inf")
        max_end = 0
        min_cds_start = float("inf")
        max_cds_end = 0
        transcript_ids = []
        transcript_names = []
        transcript_types = set()
        gene_ids = set()
        gene_names = set()
        gene_types = set()
        chroms = set()
        strands = set()
        exons = []
        for transcript in transcripts:
            transcript_ids.append(transcript.transcript_id)
            transcript_names.append(transcript.transcript_name)
            transcript_types.add(transcript.transcript_type)
            gene_ids.add(transcript.gene_id)
            gene_names.add(transcript.gene_name)
            gene_types.add(transcript.gene_type)
            chroms.add(transcript.chrom)
            strands.add(transcript.strand)
            exons.extend(transcript.exons())
            if min_start > transcript.start:
                min_start = transcript.start
            if max_end < transcript.end:
                max_end = transcript.end
            if transcript.cds_start is None and transcript.cds_end is None:
                continue
            if min_cds_start > transcript.cds_start:
                min_cds_start = transcript.cds_start
            if max_cds_end < transcript.cds_end:
                max_cds_end = transcript.cds_end

        exons = merge_intervals(set(exons))
        start = min_start
        end = max_end
        if min_cds_start == float("inf"):
            min_cds_start = None
        if max_cds_end == 0:
            max_cds_end = None
        cds_start = min_cds_start
        cds_end = max_cds_end
        count = len(transcript_ids)
        transcript_id = ",".join(transcript_ids)
        transcript_name = ",".join(transcript_names)
        transcript_type = ",".join(list(transcript_types))
        gene_id = ",".join(list(gene_ids))
        gene_name = ",".join(list(gene_names))
        gene_type = ",".join(list(gene_types))
        chrom = ",".join(list(chroms))
        strand = ",".join(list(strands))
        return cls(transcript_id, transcript_name, transcript_type,
                   gene_id, gene_name, gene_type,
                   chrom, start, end, strand,
                   cds_start, cds_end, exons, count)

    def to_bed6(self) -> Bed6:
        return Bed6(self.chrom, self.start, self.end, self.gene_name, str(self.count), self.strand)

    def upStream(self):
        if self.strand == '-':
            return self.after
        else:
            return self.before

    def downStream(self):
        if self.strand == '-':
            return self.before
        else:
            return self.after

    def length(self):
        return self.end - self.start


def group_overlap_transcripts(iterable: Iterable[Transcript]) -> Iterator[Tuple[Transcript, ...]]:
    a = deque(itertools.islice(iterable, 1))
    max_end = 0
    for i in iterable:
        before, later = a[-1], i
        max_end = max(max_end, before.end)
        if (before.chrom == later.chrom) and (later.start <= max_end):
            a.append(i)
        else:
            yield tuple(a)
            if not before.chrom == later.chrom:
                max_end = 0
            a.clear()
            a.append(i)
    if a:
        yield tuple(a)
        a.clear()


def add_before_and_after(x: Tuple[MergedTranscript, ...]):
    x[0].after = x[1].start - x[0].end
    x[1].before = x[0].after


def add_chrom_ends_before_and_after(x: MergedTranscript, chrom_length_dict: Optional[Dict[str, int]] = None):
    if x.before is None:
        x.before = x.start
    if x.after is None:
        if chrom_length_dict is not None:
            try:
                x.after = chrom_length_dict[x.chrom] - x.end
            except KeyError:
                x.after = 0
        else:
            x.after = 0


def merge_transcripts(gtf: GtfFile, strand: Optional[Literal["+", "-"]] = "+",
                      escape_gene_types: Sequence[str] = (),
                      escape_gene_name_startswith: Tuple[str] = (),
                      chrom_length_dict: Optional[Dict[str, int]] = None
                      ) -> List[MergedTranscript]:
    escape = set(escape_gene_types)
    add_chrom_ends_before_and_after_partial = partial(add_chrom_ends_before_and_after, chrom_length_dict=chrom_length_dict)
    merged_transcripts = gtf.to_transcript() | filter(lambda x: not x.gene_name.startswith(
        escape_gene_name_startswith)) | filter(lambda x: x.gene_type not in escape) | filter(
            lambda x: x.strand == strand if strand else True) | transform(
                group_overlap_transcripts) | mapwith(MergedTranscript.init_by_transcripts) | window | filter(
                    lambda x: x[0].chrom == x[-1].chrom) | apply(add_before_and_after) | flatten | uniq(
                        lambda x: x.gene_name) | apply(add_chrom_ends_before_and_after_partial)
    return merged_transcripts


def merge_transcripts_groupby_strand(
    gtf: GtfFile,
    escape_gene_types: Sequence[str] = (),
    escape_gene_name_startswith: Tuple[str] = (),
    chrom_length_dict: Optional[Dict[str, int]] = None,
    savebed: Optional[str] = None,
) -> Dict[str, List[MergedTranscript]]:
    bedpath = savebed if savebed is not None else os.devnull
    with open(bedpath, "w") as bed:
        merged_transcripts_dict = {
            '+': merge_transcripts(gtf, '+', escape_gene_types, escape_gene_name_startswith, chrom_length_dict) | apply(
                lambda x: bed.write(f"{x.to_bed6()}\n")) | to_list,
            '-': merge_transcripts(gtf, '-', escape_gene_types, escape_gene_name_startswith, chrom_length_dict) | apply(
                lambda x: bed.write(f"{x.to_bed6()}\n")) | to_list,
        }
    if savebed is not None:
        bedtools_sort(savebed)
    return merged_transcripts_dict



# -*- coding: utf-8 -*-
import re
from dataclasses import dataclass, field
from io import TextIOWrapper
from typing import List, Sequence, Tuple, Literal, Iterable, Iterator, Optional, Union, TextIO

from pybiotk.annodb import Transcript
from pybiotk.io.bed import Bed6, Bed12, Intron, GeneInfo, TransInfo
from stream.pipe import Pipe, sort, drop_while, filter, kgroupby, groupby, apply, window


@dataclass
class GTF:
    seqname: str = field(default=None)
    source: str = field(default=None, repr=False)
    feature: str = field(default=None)
    start: int = field(default=0)
    end: int = field(default=0)
    score: str = field(default=None, repr=False)
    strand: Literal['+', '-'] = field(default='+')
    frame: str = field(default=None, repr=False)
    attributes: str = field(default=None, repr=False)
    chrom: str = field(init=False, repr=False)

    def __post_init__(self):
        self.chrom = self.seqname
        self.start = int(self.start)
        self.end = int(self.end)

    def __str__(self):
        return "\t".join(str(s) for s in list(self.__dict__.values())[:9])

    @staticmethod
    def parse_attributes(term: str, attributes: str) -> Optional[str]:
        parse_attr = re.compile('{term} "([^"]+)"'.format(term=term))
        patterns = parse_attr.findall(attributes)

        if patterns:
            return patterns[0]
        else:
            return None

    def gene_type(self):
        return self.parse_attributes("gene_type", self.attributes)

    def gene_id(self):
        return self.parse_attributes("gene_id", self.attributes)

    def gene_name(self):
        return self.parse_attributes("gene_name", self.attributes)

    def transcript_type(self):
        return self.parse_attributes("transcript_type", self.attributes)

    def transcript_id(self):
        return self.parse_attributes("transcript_id", self.attributes)

    def transcript_name(self):
        return self.parse_attributes("transcript_name", self.attributes)

    def get_attribute(self, attribute):
        return self.parse_attributes(attribute, self.attributes)


@Pipe
def to_GTF(iterable: Iterable[str]) -> Iterator[GTF]:
    for line in iterable:
        yield GTF(*line.rstrip("\r\n").split("\t"))


@Pipe
def to_Bed6(iterable: Iterable[Tuple[str, Tuple[GTF, ...]]]) -> Iterator[Bed6]:
    for (k, gtfs) in iterable:
        for gtf in gtfs:
            yield Bed6.init_by_gtf(gtf, k)


@Pipe
def to_GeneInfo(iterable: Iterable[Tuple[str, Tuple[GTF, ...]]]) -> Iterator[GeneInfo]:
    for (k, gtfs) in iterable:
        for gtf in gtfs:
            yield GeneInfo.init_by_gtf(gtf, k)


@Pipe
def to_TransInfo(iterable: Iterable[Tuple[str, Tuple[GTF, ...]]]) -> Iterator[TransInfo]:
    for (k, gtfs) in iterable:
        for gtf in gtfs:
            yield TransInfo.init_by_gtf(gtf, k)


@Pipe
def to_Bed12(iterable: Iterable[Tuple[GTF, ...]], name: str = "transcript_id") -> Iterator[Bed12]:

    for group in iterable:
        gtfs: Iterator[GTF] = group | sort(key=lambda x: x.start)

        cds_exons: Sequence[GTF] = []
        bed: Optional[Bed12] = None
        for gtf in gtfs:
            if gtf.feature in {'CDS', 'stop_codon'}:
                cds_exons.append(gtf)
            elif gtf.feature == "exon":
                if bed is None:
                    bed = Bed12.init_by_gtf(gtf, gtf.get_attribute(name))
                else:
                    bed.update(gtf)
        if cds_exons:
            bed.thickStart = cds_exons[0].start - 1
            bed.thickEnd = cds_exons[-1].end
        else:
            bed.thickStart = bed.thickEnd = bed.end

        yield bed


@Pipe
def to_Intron(iterable: Iterable[Tuple[GTF, ...]], name: str = "transcript_id") -> Iterator[Intron]:
    for group in iterable:
        gtfs = group | sort(key=lambda x: x.start)
        for exons in gtfs | window:
            if len(exons) < 2:
                continue
            yield Intron.init_by_gtf(exons[0], exons[1], exons[0].get_attribute(name))


@Pipe
def to_Transcript(iterable: Iterable[Tuple[GTF, ...]]) -> Iterator[Transcript]:
    for group in iterable:
        gtfs: Iterator[GTF] = group | sort(key=lambda x: x.start)

        cds_exons: Sequence[GTF] = []
        transcript: Optional[Transcript] = None
        for gtf in gtfs:
            if gtf.feature in {'CDS', 'stop_codon'}:
                cds_exons.append(gtf)
            elif gtf.feature == "exon":
                if transcript is None:
                    transcript = Transcript.init_by_gtf(gtf)
                else:
                    transcript.update(gtf)
        if cds_exons:
            cds_start = cds_exons[0].start - 1
            cds_end = cds_exons[-1].end
            transcript.cds_start = cds_start
            transcript.cds_end = cds_end
        else:
            transcript.cds_start = transcript.cds_end = None

        yield transcript


class GtfFile:
    def __init__(self, filepath_or_buffer: Union[str, TextIO], comment: str = "#"):
        if isinstance(filepath_or_buffer, TextIOWrapper):
            self.filename: str = filepath_or_buffer.name
            self.gtf: TextIO = filepath_or_buffer
        else:
            self.filename: str = filepath_or_buffer
            self.gtf: TextIO = open(filepath_or_buffer)
        self.comment: str = comment
        self.gtf_list: Optional[Sequence[GTF]] = None
        self.ptr: int = 0
        self.closed: bool = False

    def __iter__(self) -> Iterator[GTF]:
        if self.gtf_list is not None:
            return iter(self.gtf_list)
        else:
            self.gtf_list = []
            iterator = self.gtf | drop_while(lambda x: x.startswith(self.comment)) | to_GTF | apply(lambda x: self.gtf_list.append(x))
            return iterator

    def __next__(self) -> GTF:
        if self.gtf_list is None:
            self.to_list()
        if self.ptr < len(self.gtf_list):
            gtf = self.gtf_list[self.ptr]
            self.ptr += 1
        else:
            raise StopIteration

        return gtf

    def close(self):
        if not self.closed:
            self.gtf.close()
            self.closed = True
            self.gtf_list = None
            self.ptr = 0

    def to_list(self) -> Optional[List[GTF]]:
        if self.gtf_list is None:
            _ = list(self)
        return self.gtf_list

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, traceback):
        self.close()

    def iter(self,
             features: Optional[Sequence[str]] = None,
             gene_types: Optional[Sequence[str]] = None,
             transcript_types: Optional[Sequence[str]] = None,
             transcript_ids: Optional[Sequence[str]] = None,
             transcript_names: Optional[Sequence[str]] = None,
             gene_ids: Optional[Sequence[str]] = None,
             gene_names: Optional[Sequence[str]] = None,
             ) -> Iterator[GTF]:

        iterator = iter(self)
        if features is not None:
            features = set(features)
            iterator = iterator | filter(lambda x: x.feature in features)
        if gene_types is not None:
            gene_types = set(gene_types)
            iterator = iterator | filter(
                lambda x: x.gene_type() in gene_types)
        if transcript_types is not None:
            transcript_types = set(transcript_types)
            iterator = iterator | filter(
                lambda x: x.transcript_type() in transcript_types)
        if transcript_ids is not None:
            transcript_ids = set(transcript_ids)
            iterator = iterator | filter(lambda x: x.transcript_id() in transcript_ids)
        if transcript_names is not None:
            transcript_names = set(transcript_names)
            iterator = iterator | filter(lambda x: x.transcript_name() in transcript_names)
        if gene_ids is not None:
            gene_ids = set(gene_ids)
            iterator = iterator | filter(lambda x: x.gene_id() in gene_ids)
        if gene_names is not None:
            gene_names = set(gene_names)
            iterator = iterator | filter(lambda x: x.gene_name() in gene_names)

        return iterator

    def iter_exon(self,
                  gene_types: Optional[Sequence[str]] = None,
                  transcript_types: Optional[Sequence[str]] = None,
                  transcript_ids: Optional[Sequence[str]] = None,
                  transcript_names: Optional[Sequence[str]] = None,
                  gene_ids: Optional[Sequence[str]] = None,
                  gene_names: Optional[Sequence[str]] = None,
                  ) -> Iterator[GTF]:
        return self.iter(features=["exon"],
                         gene_types=gene_types,
                         transcript_types=transcript_types,
                         transcript_ids=transcript_ids,
                         transcript_names=transcript_names,
                         gene_ids=gene_ids,
                         gene_names=gene_names
                         )

    def iter_gene(self,
                  gene_types: Optional[Sequence[str]] = None,
                  transcript_types: Optional[Sequence[str]] = None,
                  transcript_ids: Optional[Sequence[str]] = None,
                  transcript_names: Optional[Sequence[str]] = None,
                  gene_ids: Optional[Sequence[str]] = None,
                  gene_names: Optional[Sequence[str]] = None,
                  ) -> Iterator[GTF]:
        return self.iter(features=["gene"],
                         gene_types=gene_types,
                         transcript_types=transcript_types,
                         transcript_ids=transcript_ids,
                         transcript_names=transcript_names,
                         gene_ids=gene_ids,
                         gene_names=gene_names
                         )

    def iter_transcript(self,
                        gene_types: Optional[Sequence[str]] = None,
                        transcript_types: Optional[Sequence[str]] = None,
                        transcript_ids: Optional[Sequence[str]] = None,
                        transcript_names: Optional[Sequence[str]] = None,
                        gene_ids: Optional[Sequence[str]] = None,
                        gene_names: Optional[Sequence[str]] = None,
                        ) -> Iterator[GTF]:
        return self.iter(features=["transcript"],
                         gene_types=gene_types,
                         transcript_types=transcript_types,
                         transcript_ids=transcript_ids,
                         transcript_names=transcript_names,
                         gene_ids=gene_ids,
                         gene_names=gene_names
                         )

    def to_bed6(self, feature: Literal["exon", "gene"] = "gene",
                name: Literal["gene_id", "gene_name", "transcript_id", "transcript_name"] = "gene_name",
                gene_types: Optional[Sequence[str]] = None,
                transcript_types: Optional[Sequence[str]] = None,
                transcript_ids: Optional[Sequence[str]] = None,
                transcript_names: Optional[Sequence[str]] = None,
                gene_ids: Optional[Sequence[str]] = None,
                gene_names: Optional[Sequence[str]] = None
                ) -> Iterator[Bed6]:

        bed6 = self.iter(features=[feature],
                         gene_types=gene_types,
                         transcript_types=transcript_types,
                         transcript_ids=transcript_ids,
                         transcript_names=transcript_names,
                         gene_ids=gene_ids,
                         gene_names=gene_names
                         ) | kgroupby(lambda x: x.get_attribute(name)) | to_Bed6
        return bed6

    def to_gene_info(self,
                     gene_types: Optional[Sequence[str]] = None,
                     transcript_types: Optional[Sequence[str]] = None,
                     transcript_ids: Optional[Sequence[str]] = None,
                     transcript_names: Optional[Sequence[str]] = None,
                     gene_ids: Optional[Sequence[str]] = None,
                     gene_names: Optional[Sequence[str]] = None,
                     ) -> Iterator[GeneInfo]:
        gene_info = self.iter_gene(
            gene_types=gene_types,
            transcript_types=transcript_types,
            transcript_ids=transcript_ids,
            transcript_names=transcript_names,
            gene_ids=gene_ids,
            gene_names=gene_names
            ) | kgroupby(lambda x: x.gene_id()) | to_GeneInfo
        return gene_info

    def to_trans_info(self,
                      gene_types: Optional[Sequence[str]] = None,
                      transcript_types: Optional[Sequence[str]] = None,
                      transcript_ids: Optional[Sequence[str]] = None,
                      transcript_names: Optional[Sequence[str]] = None,
                      gene_ids: Optional[Sequence[str]] = None,
                      gene_names: Optional[Sequence[str]] = None,
                      ) -> Iterator[TransInfo]:
        trans_info = self.iter_transcript(
            gene_types=gene_types,
            transcript_types=transcript_types,
            transcript_ids=transcript_ids,
            transcript_names=transcript_names,
            gene_ids=gene_ids,
            gene_names=gene_names
            ) | kgroupby(lambda x: x.transcript_id()) | to_TransInfo
        return trans_info

    def to_bed12(self, name: Literal["gene_id", "gene_name", "transcript_id", "transcript_name"] = "transcript_id",
                 gene_types: Optional[Sequence[str]] = None,
                 transcript_types: Optional[Sequence[str]] = None,
                 transcript_ids: Optional[Sequence[str]] = None,
                 transcript_names: Optional[Sequence[str]] = None,
                 gene_ids: Optional[Sequence[str]] = None,
                 gene_names: Optional[Sequence[str]] = None,
                 ) -> Iterator[Bed12]:
        bed12 = self.iter(features=["CDS", 'stop_codon', 'exon'],
                          gene_types=gene_types,
                          transcript_types=transcript_types,
                          transcript_ids=transcript_ids,
                          transcript_names=transcript_names,
                          gene_ids=gene_ids,
                          gene_names=gene_names
                          ) | groupby(lambda x: x.transcript_id()) | to_Bed12(name=name)
        return bed12

    def to_intron(self, name: Literal["gene_id", "gene_name", "transcript_id", "transcript_name"] = "transcript_id",
                  gene_types: Optional[Sequence[str]] = None,
                  transcript_types: Optional[Sequence[str]] = None,
                  transcript_ids: Optional[Sequence[str]] = None,
                  transcript_names: Optional[Sequence[str]] = None,
                  gene_ids: Optional[Sequence[str]] = None,
                  gene_names: Optional[Sequence[str]] = None,
                  ) -> Iterator[Intron]:
        intron = self.iter_exon(
            gene_types=gene_types,
            transcript_types=transcript_types,
            transcript_ids=transcript_ids,
            transcript_names=transcript_names,
            gene_ids=gene_ids,
            gene_names=gene_names
            ) | groupby(lambda x: x.transcript_id()) | to_Intron(name=name)
        return intron

    def to_transcript(self,
                      gene_types: Optional[Sequence[str]] = None,
                      transcript_types: Optional[Sequence[str]] = None,
                      transcript_ids: Optional[Sequence[str]] = None,
                      transcript_names: Optional[Sequence[str]] = None,
                      gene_ids: Optional[Sequence[str]] = None,
                      gene_names: Optional[Sequence[str]] = None,
                      ) -> Iterator[Transcript]:
        transcript = self.iter(features=["CDS", 'stop_codon', 'exon'],
                               gene_types=gene_types,
                               transcript_types=transcript_types,
                               transcript_ids=transcript_ids,
                               transcript_names=transcript_names,
                               gene_ids=gene_ids,
                               gene_names=gene_names
                               ) | groupby(lambda x: x.transcript_id()) | to_Transcript
        return transcript



#!/usr/bin/env python3
import argparse
import os
import sys
import time
from collections import Counter

from pybiotk.intervals import merge_intervals
from pybiotk.io import Bam, BamPE, BamType, check_bam_type
from pybiotk.utils import logging, blocks_len, intervals_is_overlap, ignore


def main(filename: str, output: str, ordered_by_name: bool = False):
    start = time.perf_counter()
    bamtype = check_bam_type(filename)
    length = []
    sys.stdout.write("read_name\treference_name\tblocks\tlength\n")
    if bamtype is BamType.SE:
        logging.info("SingleEnd mode ...")
        with Bam(filename) as bam:
            for read in bam.iter_mapped():
                blocks = read.get_blocks()
                read_len = blocks_len(blocks)
                length.append(read_len)
                sys.stdout.write(f"{read.query_name}\t{read.reference_name}\t{blocks}\t{read_len}\n")
    elif bamtype is BamType.PE:
        logging.info("PairEnd mode ...")
        with BamPE(filename) as bampe:
            bampe.ordered_by_name = ordered_by_name
            for read1, read2 in bampe.iter_pair(properly_paired=False):
                if read1 is not None and read2 is not None:
                    if read1.reference_name == read2.reference_name:
                        read1_blocks = read1.get_blocks()
                        read2_blocks = read2.get_blocks()
                        if not intervals_is_overlap(read1_blocks, read2_blocks):
                            logging.warning(f"{read1.query_name} read1 and read2 not overlap, use read1_len + read2_len")
                        merge_blocks = merge_intervals(read1_blocks+read2_blocks)
                        fragment_len = blocks_len(merge_blocks)
                        length.append(fragment_len)
                        sys.stdout.write(f"{read1.query_name}\t{read1.reference_name}\t{merge_blocks}\t{fragment_len}")
                    elif not read1.reference_name == read2.reference_name:
                        read1_blocks = read1.get_blocks()
                        read1_len = blocks_len(read1_blocks)
                        length.append(read1_len)
                        sys.stdout.write(f"{read1.query_name}_R1\t{read1.reference_name}\t{read1_blocks}\t{read1_len}\n")
                        read2_blocks = read2.get_blocks()
                        read2_len = blocks_len(read2_blocks)
                        length.append(read2_len)
                        sys.stdout.write(f"{read2.query_name}_R2\t{read2.reference_name}\t{read2_blocks}\t{read2_len}\n")
                if read1 is not None and read2 is None:
                    read1_blocks = read1.get_blocks()
                    read1_len = blocks_len(read1_blocks)
                    length.append(read1_len)
                    sys.stdout.write(f"{read1.query_name}_R1\t{read1.reference_name}\t{read1_blocks}\t{read1_len}\n")
                if read2 is not None and read1 is None:
                    read2_blocks = read2.get_blocks()
                    read2_len = blocks_len(read2_blocks)
                    length.append(read2_len)
                    sys.stdout.write(f"{read2.query_name}_R2\t{read2.reference_name}\t{read2_blocks}\t{read2_len}\n")
    counter = Counter(length)
    counter = list(counter.items())
    counter.sort(key=lambda x: x[0])
    with open(output, "w") as f:
        f.write("Length\tCount\n")
        for length, count in counter:
            f.write(f"{length}\t{count}\n")

    end = time.perf_counter()
    logging.info(f"task finished in {end-start:.2f}s.")


@ignore
def run():
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("input", type=str, help="input bam.")
    parser.add_argument("-s", dest="summary", type=str, default=os.devnull,
                        help="output summary file name.")
    parser.add_argument("--ordered_by_name", dest="ordered_by_name", action="store_true",
                        help="if input bam is ordered by name.")

    args = parser.parse_args()
    main(args.input, args.summary, args.ordered_by_name)


if __name__ == "__main__":
    run()



# -*- coding: utf-8 -*-
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import List, Tuple, Set, Iterable, AbstractSet, Optional

from pybiotk.utils import blocks_len, intervals_is_overlap


@dataclass
class GenomicAnnotation:
    id: Optional[str] = field(default=None)
    name: Optional[str] = field(default=None)
    start: Optional[int] = field(default=None)
    end: Optional[int] = field(default=None)
    type: Optional[str] = field(default=None)
    detail: Set[str] = field(default_factory=set)

    def update(self, anno: str):
        self.detail.add(anno)

    @staticmethod
    def select_anno(
        annoset: AbstractSet[str],
        priority: Tuple[str, ...] = ("Promoter", "5UTR", "3UTR", "CDS", "Exon", "Intron", "Downstream", "Intergenic")
    ):
        for anno in priority:
            if anno in annoset:
                return anno

    def primary_anno(
        self,
        priority: Tuple[str, ...] = ("Promoter", "5UTR", "3UTR", "CDS", "Exon", "Intron", "Downstream", "Intergenic")
    ):
        return self.select_anno(self.detail, priority)


@dataclass
class AnnoSet:
    annoset: Iterable[GenomicAnnotation] = field(default_factory=list, repr=False)
    id: List[str] = field(init=False, default_factory=list)
    name: List[str] = field(init=False, default_factory=list)
    start: List[int] = field(init=False, default_factory=list)
    end: List[int] = field(init=False, default_factory=list)
    type: List[str] = field(init=False, default_factory=list)
    anno: List[str] = field(init=False, default_factory=list)

    def __post_init__(self):
        for anno in self.annoset:
            self.id.append(anno.id)
            self.name.append(anno.name)
            self.start.append(anno.start)
            self.end.append(anno.end)
            self.type.append(anno.type)
            self.anno.append(anno.primary_anno())

    def primary_anno(
        self,
        priority: Tuple[str, ...] = ("Promoter", "5UTR", "3UTR", "CDS", "Exon", "Intron", "Downstream", "Intergenic")
    ) -> str:
        return GenomicAnnotation.select_anno(set(self.anno), priority)

    def __str__(self) -> str:
        _id = ",".join(self.id)
        _name = ",".join(set(self.name))
        _type = ",".join(set(self.type))
        _start = ",".join(str(i) for i in set(self.start))
        _end = ",".join(str(i) for i in set(self.end))
        _anno = self.primary_anno()
        return f"{_anno}\t{_start}\t{_end}\t{_name}\t{_id}\t{_type}"


class GFeature(ABC):
    @abstractmethod
    def is_protein_coding(self) -> bool: ...

    @abstractmethod
    def exons(self) -> List[Tuple[int, int]]: ...

    @abstractmethod
    def introns(self) -> List[Tuple[int, int]]: ...

    @abstractmethod
    def tss_region(self, region: Tuple[int, int] = (-1000, 1000)) -> Tuple[int, int]: ...

    @abstractmethod
    def downstream(self, down: int = 3000) -> Tuple[int, int]: ...

    @abstractmethod
    def cds_exons(self) -> List[Tuple[int, int]]: ...

    @abstractmethod
    def utr5_exons(self) -> List[Tuple[int, int]]: ...

    @abstractmethod
    def utr3_exons(self) -> List[Tuple[int, int]]: ...

    def exons_len(self) -> int:
        return blocks_len(self.exons())

    def introns_len(self) -> int:
        return blocks_len(self.introns())

    def cds_len(self) -> int:
        return blocks_len(self.cds_exons())

    def utr5_len(self) -> int:
        return blocks_len(self.utr5_exons())

    def utr3_len(self) -> int:
        return blocks_len(self.utr3_exons())

    def anno(self, blocks: List[Tuple[int, int]], region: Tuple[int, int] = (-1000, 1000), down: int = 3000) -> Set[str]:
        anno = []
        tss = intervals_is_overlap(blocks, [self.tss_region(region=region)])
        introns = self.introns()
        if introns:
            introns = intervals_is_overlap(blocks, introns)
        downstream = intervals_is_overlap(blocks, [self.downstream(down=down)])
        if self.is_protein_coding():
            utr5_exons = self.utr5_exons()
            if utr5_exons:
                utr5_exons = intervals_is_overlap(blocks, utr5_exons)
            utr3_exons = self.utr3_exons()
            if utr3_exons:
                utr3_exons = intervals_is_overlap(blocks, utr3_exons)
            cds_exons = self.cds_exons()
            if cds_exons:
                cds_exons = intervals_is_overlap(blocks, cds_exons)
            exons = False
        else:
            exons = intervals_is_overlap(blocks, self.exons())
            utr5_exons = False
            utr3_exons = False
            cds_exons = False

        if tss:
            anno.append("Promoter")
        if self.is_protein_coding():
            if utr5_exons:
                anno.append("5UTR")
            if utr3_exons:
                anno.append("3UTR")
            if cds_exons:
                anno.append("CDS")
        elif exons:
            anno.append("Exon")
        if introns:
            anno.append("Intron")
        if downstream:
            anno.append("Downstream")
        if not anno:
            anno.append("Intergenic")
        return set(anno)



import argparse
from collections import defaultdict

from pybiotk.io import Bam, BamPE, BamType, check_bam_type


def main(filename: str, output: str, count_fragments: bool = False, ordered_by_name: bool = False):
    reference_count_dict = defaultdict(int)
    bamtype = check_bam_type(filename)
    if bamtype is BamType.PE and count_fragments:
        with BamPE(filename, ordered_by_name=ordered_by_name) as bam:
            for read1, read2 in bam.iter_pair(properly_paired=False):
                if read1 is not None and read2 is not None:
                    if read1.reference_name == read2.reference_name:
                        reference_count_dict[read1.reference_name] += 1
                    elif not read1.reference_name == read2.reference_name:
                        reference_count_dict[read1.reference_name] += 1
                        reference_count_dict[read2.reference_name] += 1
                if read1 is not None and read2 is None:
                    reference_count_dict[read1.reference_name] += 1
                if read2 is not None and read1 is None:
                    reference_count_dict[read2.reference_name] += 1
    else:
        with Bam(filename) as bam:
            for read in bam.iter_mapped():
                reference_count_dict[read.reference_name] += 1

    with open(output, 'w') as outfile:
        outfile.write(f"Geneid\t{filename}\n")
        for reference in reference_count_dict:
            outfile.write(f"{reference}\t{reference_count_dict[reference]}\n")


def run():
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("input", type=str, help="input bam.")
    parser.add_argument("-o", dest="output", type=str, required=True,
                        help="output file.")
    parser.add_argument("-p", "--pair", dest="pair", action="store_true",
                        help="count fragments instead of reads.")
    parser.add_argument("--ordered_by_name", dest="ordered_by_name", action="store_true",
                        help="if input bam is ordered by name.")

    args = parser.parse_args()
    main(args.input, args.output, args.pair, args.ordered_by_name)



#!/usr/bin/env python3
import argparse
import sys
from typing import Literal, Sequence, Optional, TextIO

from pybiotk.io import GtfFile
from pybiotk.utils import ignore
from stream import write


def main(
        istream: TextIO,
        name: Literal["gene_id", "gene_name", "transcript_id",
                      "transcript_name"] = "transcript_id",
        feature: Literal["exon", "gene"] = "exon",
        gene_types: Optional[Sequence[str]] = None,
        transcript_types: Optional[Sequence[str]] = None,
        transcript_ids: Optional[Sequence[str]] = None,
        transcript_names: Optional[Sequence[str]] = None,
        gene_ids: Optional[Sequence[str]] = None,
        gene_names: Optional[Sequence[str]] = None,
        output: TextIO = sys.stdout,
        outformat: Literal["bed12", "bed6", "intron", "gene_info", "trans_info"] = "bed12"):
    with GtfFile(istream) as gtf:
        if outformat == 'bed12':
            gtf.to_bed12(name, gene_types, transcript_types, transcript_ids, transcript_names, gene_ids, gene_names) | write(output)
        elif outformat == 'bed6':
            gtf.to_bed6(feature, name, gene_types, transcript_types, transcript_ids, transcript_names, gene_ids, gene_names) | write(output)
        elif outformat == 'intron':
            gtf.to_intron(name, gene_types, transcript_types, transcript_ids, transcript_names, gene_ids, gene_names) | write(output)
        elif outformat == 'gene_info':
            gtf.to_gene_info(gene_types, transcript_types, transcript_ids, transcript_names, gene_ids, gene_names) | write(output)
        elif outformat == 'trans_info':
            gtf.to_trans_info(gene_types, transcript_types, transcript_ids, transcript_names, gene_ids, gene_names) | write(output)
        else:
            raise RuntimeError(f"{outformat} not support.")


@ignore
def run():
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('gtf', nargs='?', type=argparse.FileType('r'),
                        default=(None if sys.stdin.isatty() else sys.stdin),
                        help="gtf file download from Genecode, or a sorted gtf file.")
    parser.add_argument('-n', '--name', dest='name', default='transcript_id',
                        choices=["gene_id", "gene_name", "transcript_id", "transcript_name"],
                        help="""attribute ID by which to group bed entries.
                                'gene_id' or 'gene_name' for gene.bed6, gene_info,
                                'transcript_id' for genome.bed12, exon, intron, transcrpt_info"""
                        )
    parser.add_argument('-f', '--feature', dest='feature', default='exon',
                        choices=["exon", "gene"],
                        help="""
                        annotation feature to join, all others are filtered out:
                        'exon' genome.bed12, exon, intron, 'gene' for gene.bed6, gene_info."""
                        )
    parser.add_argument('-o', dest='output', type=argparse.FileType('w'),
                        default=sys.stdout, help="output file name.")
    parser.add_argument('--outformat', dest='outformat', default='bed12',
                        choices=["bed12", "bed6", "intron", "gene_info", "trans_info"],
                        help="choose output file format:bed12, bed6, intron, info")
    parser.add_argument('--gene_types', dest='gene_types', nargs="+",
                        default=None, help="choose gene types to filter bed.")
    parser.add_argument('--transcript_types', dest='transcript_types', nargs="+",
                        default=None, help="choose transcript types to filter bed.")
    parser.add_argument('--transcript_ids', dest='transcript_ids', nargs="+",
                        default=None, help="choose transcript ids to filter bed.")
    parser.add_argument('--transcript_names', dest='transcript_names', nargs="+",
                        default=None, help="choose transcript names to filter bed.")
    parser.add_argument('--gene_ids', dest='gene_ids', nargs="+",
                        default=None, help="choose gene ids to filter bed.")
    parser.add_argument('--gene_names', dest='gene_names', nargs="+",
                        default=None, help="choose gene names to filter bed.")

    args = parser.parse_args()
    if args.gtf is None:
        parser.parse_args(['-h'])

    main(args.gtf, args.name, args.feature,
         args.gene_types, args.transcript_types,
         args.transcript_ids, args.transcript_names,
         args.gene_ids, args.gene_names,
         args.output, args.outformat)


if __name__ == "__main__":
    run()



#!/usr/bin/env python3
import argparse
import sys

import pysam


def merge_sub(seq1, seq2):
    if seq1.sequence.find(seq2.sequence) != -1:
        seq1.name = seq1.name + ";" + seq2.name
        return seq1
    elif seq2.sequence.find(seq1.sequence) != -1:
        seq2.name = seq1.name + ";" + seq2.name
        return seq2
    else:
        return None


def overlaps(a):
    i = 0
    while i < len(a):
        j = 0
        while j < len(a) and i < len(a):
            if i == j:
                j += 1
                continue
            if a[i].sequence.find(a[j].sequence) != -1:
                a[i].name = a[i].name + ";" + a[j].name
                a.pop(j)
                j -= 1
            elif a[j].sequence.find(a[i].sequence) != -1:
                a[j].name = a[j].name + ";" + a[j].name
                a[i] = a[j]
                a.pop(j)
                j -= 1
            j += 1
        i += 1
    return a


def run_method1(inputfa):
    merge_list = []
    with pysam.FastxFile(inputfa) as fh:
        i = 0
        for entry in fh:
            i += 1
            sys.stderr.write(f"\rprocess {i} reads")
            if not merge_list:
                merge_list.append(entry)
                continue
            collapse_seq = entry
            filter_seq_list = []
            for i, seq in enumerate(merge_list):
                merge_seq = merge_sub(seq, collapse_seq)
                if merge_seq:
                    collapse_seq = merge_seq
                else:
                    filter_seq_list.append(merge_list[i])

            merge_list = filter_seq_list
            merge_list.append(collapse_seq)
        sys.stderr.write("\n")
        for i in merge_list:
            print(f">{i.name}\n{i.sequence}")


def run_method2(inputfa):
    with pysam.FastxFile(inputfa) as fh:
        results = overlaps(list(fh))

    for i in results:
        print(f">{i.name}\n{i.sequence}")


def run():
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument(dest="fasta", type=str, help="fasta")
    parser.add_argument("-m", dest="method", type=int, default=1, help="method1 or method2")

    args = parser.parse_args()
    if not args.method == 1:
        run_method2(args.fasta)
    else:
        run_method1(args.fasta)


if __name__ == "__main__":
    run()



#!/usr/bin/env python3
import argparse
import os
import sys
from collections import defaultdict
from itertools import groupby

import pandas as pd

from pybiotk.utils import split_discontinuous, ignore


def main(bed_list, header):
    if header:
        name = os.path.splitext(os.path.basename(bed_list[0]))[0]
        print(f'track type=bedGraph name="{name}" visibility=full color=0,51,51 AutoScale=on alwaysZero=on maxHeightPixels=50:50:50')
    df_list = []
    for bedfile in bed_list:
        df_list.append(pd.read_table(bedfile, header=None))
    df = pd.concat(df_list)
    bed = df.sort_values(by=0).reset_index(drop=True)
    for chrom, group in groupby(bed.iterrows(), lambda x: x[1][0]):
        values = defaultdict(int)
        for _, row in group:
            start = row[1]
            end = row[2]
            for pos in range(start, end):
                values[int(pos)] += 1
        values_items = sorted(values.items(), key=lambda x: x[0])
        for value, g in groupby(values_items, lambda x: x[1]):
            pos = [i[0] for i in g]
            for pos_list in split_discontinuous(pos):
                sys.stdout.write(f"{chrom}\t{pos_list[0]}\t{pos_list[-1]}\t{value}\n")


@ignore
def run():
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("input", type=str, nargs="+",
                        help="Input bedfiles")
    parser.add_argument("--header", dest="header", action="store_true",
                        help="Add a header to the output file")
    args = parser.parse_args()
    main(args.input, args.header)


if __name__ == "__main__":
    run()
